from tkinter import Canvas, Frame, Label

import numpy as np

from .tool_tip import ToolTip
from .labels import NumberEditLabel
from .canvas_items import CanvasEditLine, CanvasEditFill, brighten

__all__ = ['Slider', 'TimeSlider', 'HorizontalSlider', 'VerticalSlider', 'ScrollBar',
           'HorizontalSliderGroup', 'VerticalSliderGroup', 'PlotScrollBar', 'DoubleScrollBar', 'brighten']

# helper functions
def seconds_text(sec:float):
    '''converts seconds to text in form (hh):(m)m:ss'''
    sec = int(sec)
    if sec >= 3600: # at least one hour
        return f'{sec // 3600}:{(sec % 3600) // 60:0>2}:{(sec % 60) // 1:0>2}'
    else: # less than one hour
        return f'{sec // 60}:{(sec % 60) // 1:0>2}'


class Slider(Canvas):
    ''' Basic slider that takes values between 0 and 1

        Can be configured horizontally or vertically with circle or rectangle
        slider.
    '''
    def __init__(self, master:Frame, callback=None, default_value=0, bg='#000000',
                 line_color='#555555', active_line_color=None, active_line_hover_color=None,
                 slider_color='#ffffff', slider_drag_color=None, hide_slider=False,
                 slider_visible=True, slider_height=20, slider_width=20, slider_type='circle',
                 orientation='h', line_width=2, height=0, width=0, **kwargs):
        '''
        Parameters
        ----------
            :param master: tk.Frame - parent widget
            :param callback: function (float) - called only when slider is moved by user
            :param default_value: float between 0 and 1 - initial slider value
            :param bg: str (hex code) - background color
            :param line_color: str (hex code) - color of line
            :param active_line_color: str (hex code) - color of active portion of line (if different from line_color)
            :param active_line_hover_color: str (hex code) - color of active portion of line when hovering
                                                           - if different from active_line_color
            :param slider_color: str (hex code) - slider color
            :param slider_drag_color: str (hex code) - slider color when mouse is depressed (if different from slider color)
            :param hide_slider: bool - if True, slider is hidden when cursor is not hovering
            :param slider_visible: bool - if False, slider is never visible
            :param slider_height: int - height of slider in pixels (or diameter for circle slider)
            :param slider_width: int - width of slider in pixels
            :param slider_type: Literal['circle', 'rectangle'] - type of slider
            :param orientation: Literal['h', 'v'] - horizontal or vertical orientation
            :param line_width: int - width of slider line in pixels
            :param height: int - canvas height in pixels
            :param width: int - canvas width in pixels
        '''
        assert slider_type in ['circle', 'rectangle'], f'Slider Error: Invalid slider type: {slider_type}'
        assert orientation in ['h', 'v'], f'Slider Error: Invalid orientation: {orientation}'
        assert default_value >= 0 and default_value <= 1, f'Default value must be between 0 and 1, not {default_value}'
        if slider_type == 'circle' and slider_height != slider_width: # not a circle (not fatal)
            print(f'Slider Warning: Initiated circle slider but slider height ({slider_height}) does not equal slider width ({slider_width})')
        canvas_height = slider_height if orientation == 'h' else height
        canvas_width = slider_width if orientation == 'v' else width
        super().__init__(master, bg=bg, highlightthickness=0,
                         height=canvas_height, width=canvas_width, **kwargs)
        self.__value = default_value
        self.__callback = callback
        self.__orientation = orientation
        self.__slider_type = slider_type
        self.__slider_height, self.__slider_width = slider_height, slider_width
        self.__active_line_color = active_line_color if active_line_color is not None else line_color
        self.__hover_line_color = active_line_hover_color if active_line_hover_color is not None else self.__active_line_color
        self.__slider_color = slider_color if slider_color is not None else self.__active_line_color
        self.__slider_drag_color = slider_drag_color if slider_drag_color is not None else self.__slider_color
        self.__slider_state = 'hidden' if hide_slider or not slider_visible else 'disabled' # when not hovering
        self.__slider_hover_state = 'disabled' if slider_visible else 'hidden' # when hovering
        self.__dragging = False

        # Draw Lines and Slider
        self.__line_id = self.create_line(0, 0, 0, 0, fill=line_color,
                                          width=line_width, state='disabled')
        self.__active_line_id = self.create_line(0, 0, 0, 0, fill=self.__active_line_color,
                                                 width=line_width, state='disabled')
        if self.__slider_type == 'circle':
            self.__slider_id = self.create_oval(0, 0, 0, 0, fill=self.__slider_color,
                                                state=self.__slider_state)
        elif self.__slider_type == 'rectangle':
            self.__slider_id = self.create_rectangle(0, 0, 0, 0, fill=self.__slider_color,
                                                     width=0, state=self.__slider_state)
            
        # Event Bindings
        self.bind("<Button-1>", self.__click)
        self.bind("<ButtonRelease-1>", self.__release)
        self.bind("<B1-Motion>", self.__motion)
        self.bind("<Enter>", self.__hover_enter)
        self.bind("<Leave>", self.__hover_leave)
        self.bind('<Configure>', self.__set_coords)
        
        self.event_generate('<Configure>', when='tail')

    def __set_coords(self, event=None):
        '''updates coordinates of lines and slider based on canvas size and current value'''
        h, w = self.winfo_height(), self.winfo_width()
        rx, ry = self.__slider_width / 2, self.__slider_height / 2
        if self.__orientation == 'h': # horizontal
            x = (w - self.__slider_width) * self.__value + rx
            y = h / 2
            self.coords(self.__line_id, rx, y, w - rx, y)
            self.coords(self.__active_line_id, rx, y, x, y)
        elif self.__orientation == 'v': # vertical
            x = w / 2
            y = h - ry - (h - self.__slider_height) * self.__value
            self.coords(self.__line_id, w / 2, ry, w / 2, h - ry)
            self.coords(self.__active_line_id, x, y, x, h - ry)
        self.coords(self.__slider_id, x - rx, y - ry, x + rx, y + ry)

    def __get_value(self, x:float, y:float):
        '''computes value based on (x, y) coordinates'''
        if self.__orientation == 'h': # horizontal
            value = (x - self.__slider_width / 2) / (self.winfo_width() - self.__slider_width)
        elif self.__orientation == 'v': # vertical
            value = (self.winfo_height() - self.__slider_height / 2 - y) / (self.winfo_height() - self.__slider_height)
        return max(0, min(1, value))
    
    def __hover_enter(self, event=None):
        '''cursor enters canvas'''
        self.itemconfig(self.__active_line_id, fill=self.__hover_line_color)
        self.itemconfig(self.__slider_id, state=self.__slider_hover_state)

    def __hover_leave(self, event=None):
        '''cursor leaves canvas'''
        if not self.__dragging:
            self.itemconfig(self.__active_line_id, fill=self.__active_line_color)
            self.itemconfig(self.__slider_id, state=self.__slider_state)

    def __click(self, event):
        '''cursor clicks on canvas - move slider to click position'''
        self.__dragging = True
        self.itemconfig(self.__slider_id, fill=self.__slider_drag_color)
        self.__value = self.__get_value(event.x, event.y)
        self.__set_coords()
        if self.__callback is not None:
            self.__callback(self.__value)

    def __release(self, event):
        '''cursor releases click - change slider color to normal'''
        self.__dragging = False
        self.itemconfig(self.__slider_id, fill=self.__slider_color)
        w, h = self.winfo_width(), self.winfo_height()
        if event.x < 0 or event.x > w or event.y < 0 or event.y > h: # not hovering
            self.__hover_leave()

    def __motion(self, event):
        '''cursor drags slider'''
        self.__value = self.__get_value(event.x, event.y)
        self.__set_coords()
        if self.__callback is not None:
            self.__callback(self.__value)

    def set(self, value:float):
        '''sets slider value - must be between 0 and 1 (does not call callback function)'''
        assert value >= 0 and value <= 1, f'Slider Error: set slider to invalid value: {value}'
        self.__value = value
        self.__set_coords()

    def get(self):
        '''returns current slider value'''
        return self.__value

class TimeSlider(Frame):
    ''' Combines Slider with two labels to display a time slider.
    
        There can be multiple steps per second. The callback function is given
        the current step not the current time
    '''
    def __init__(self, master:Frame, callback=None, frame_num=100, start_frame=0,
                 steps_per_sec=1, bg='#000000', fg='#888888',
                 val_font_name='Segoe UI', limit_font_name='Segoe UI',
                 val_font_size=12, limit_font_size=12, **kwargs):
        '''
        Parameters
        ----------
            :param master: tk.Frame - parent widget
            :param callback: function (int) - called when slider is moved
            kwargs are passed to Slider (not Frame)
        '''
        super().__init__(master, bg=bg)
        self.__callback = callback
        self.__frame_num = frame_num
        self.__steps_per_sec = steps_per_sec

        # End Time Label - Right
        self.__end_label = Label(self, text=seconds_text(frame_num / steps_per_sec),
                                 bg=bg, fg=fg, font=(limit_font_name, limit_font_size))
        self.__end_label.pack(side='right')

        # Current Time Label - Left
        self.__label = Label(self, text=seconds_text(start_frame / steps_per_sec),
                             bg=bg, fg=fg, font=(val_font_name, val_font_size))
        self.__label.pack(side='left')

        # Slider - Center
        self.__Slider = Slider(self, self.__slider_callback,
                               default_value=start_frame / frame_num, bg=bg,
                               orientation='h', **kwargs)
        self.__Slider.pack(side='left', fill='x', expand=True)

    def __slider_callback(self, perc:float):
        '''called when slider is moved by user
        :param perc: float between 0 and 1 - new slider value
        '''
        frame = self.__frame_num * perc
        self.__label.config(text=seconds_text(frame / self.__steps_per_sec))
        if self.__callback is not None:
            self.__callback(int(frame))

    def set_frame(self, frame:int):
        '''sets current step - does not call callback function'''
        assert frame <= self.__frame_num, f'TimeSlider Error: Tried to set frame to {frame} which is greater than max frame: {self.__frame_num}'
        self.__Slider.set(frame / self.__frame_num)
        self.__label.config(text=seconds_text(frame / self.__steps_per_sec))

    def set_frame_num(self, frame_num:int, frame_rate):
        '''sets max frame - frame_rate is just there for consistency with Scrollbars'''
        self.__Slider.set(self.__frame_num * self.__Slider.get() / frame_num)
        self.__frame_num = frame_num
        self.__end_label.config(text=seconds_text(self.__frame_num / self.__steps_per_sec))

    def set_steps_per_sec(self, steps_per_sec:float):
        '''updates steps per sec without changing the number of steps'''
        self.__steps_per_sec = steps_per_sec
        self.__label.config(text=seconds_text(self.__Slider.get() * self.__frame_num / self.__steps_per_sec))
        self.__end_label.config(text=seconds_text(self.__frame_num / self.__steps_per_sec))

    def get(self) -> int:
        '''returns the current frame'''
        return int(self.__frame_num * self.__Slider.get())
    
    def get_sec(self) -> float:
        '''returns current slider position in seconds'''
        return self.__frame_num * self.__Slider.get() / self.__steps_per_sec

class LabelSlider(Frame):
    ''' Combination of Slider and NumberEditLabel so that slider value can be
        seen by the user.

        LabelSlider can handle any range of values, unlike Slider, which only
        handles values between 0 and 1.

        LabelSlider has three components: slider, number label, and text label.
        The number label displays the current value and text label is static.

        DO NOT user LabelSlider directly. It is only for inheritance from
        HorizontalSlider and VerticalSlider.
    '''
    def __init__(self, master:Frame, child_frame=None, callback=None, bg='#ffffff',
                 label:str='', popup_label=None, font_name='Segoe UI', font_size=10,
                 text_fg='#000000', popup_bg='#000000', popup_font_name=None,
                 popup_font_size=None, min_value=0, max_value=100, step=1, default_value=None,
                 label_editable=True, label_draggable=False, reference_width=2.0,
                 max_len=None, drag_threshold=0.2, label_justify='left', label_bg=None,
                 label_fg=None, label_hover_bg:str=None, label_error_color='#ff0000',
                 line_color='#555555', active_line_color=None, active_line_hover_color=None,
                 slider_color='#ffffff', slider_drag_color=None, hide_slider=False,
                 slider_visible=True, slider_height=20, slider_width=20, slider_type='circle',
                 orientation='h', line_width=2, canvas_height=0, canvas_width=0,
                 entry_on_function=None, entry_off_function=None, **kwargs):
        '''
        Parameters
        ----------
            :param master: tk.Frame - parent widget
            :param child_frame: tk.Frame - to optionally put widgets in a frame other than self
            :param callback: function (int|float), called when value is changed

        Slider Parameters
        -----------------
            :param line_color: str (hex code) - color of line
            :param active_line_color: str (hex code) - color of active portion of line (if different from line_color)
            :param active_line_hover_color: str (hex code) - color of active portion of line when hovering
                                                           - if different from active_line_color
            :param slider_color: str (hex code) - slider color
            :param slider_drag_color: str (hex code) - slider color when mouse is depressed (if different from slider color)
            :param hide_slider: bool - if True, slider is hidden when cursor is not hovering
            :param slider_visible: bool - if False, slider is never visible
            :param slider_height: int - height of slider in pixels (or diameter for circle slider)
            :param slider_width: int - width of slider in pixels
            :param slider_type: Literal['circle', 'rectangle'] - type of slider
            :param orientation: Literal['h', 'v'] - horizontal or vertical orientation
            :param line_width: int - width of slider line in pixels
            :param canvas_height: int - slider canvas height in pixels
            :param canvas_width: int - slider canvas width in pixels

        Label Parameters
        ----------
            :param min_value: int or float - minimum value
            :param max_value: int or float - maximum value
            :param step: int or float - number increment
            :param default_value: int or float - default value if different from min_value
            :param label_editable: bool - if True, label can be double clicked to enter exact value
            :param label_draggable: bool - if True, label can be dragged to adjust value
            :param reference_width: float - fraction of label width
            :param max_len: int - maximum number of characters in entry box
            :param drag_threshold: float (seconds) - click duration to be considered a single click (not drag)
            :param label_justify: Literal['left', 'right', 'center'] - justification in label entry box
            :param label_bg: str (hex code) - label background color (if different from bg)
            :param label_fg: str (hex code) - label text color (if different from text_fg)
            :param label_hover_bg: str (hex code) - label background color when hovering (if different from label_bg)
            :param label_error_color: str (hex code) - background color in entry box for bad text
            :param entry_on_function: function() - called when user opens entry box
            :param entry_of_function: function() - called when user closes entry box
        '''
        if child_frame is None: # dont need to init Frame if nothing is being put in it
            super().__init__(master, bg=bg, **kwargs)
        self.__callback = callback

        child_frame = child_frame if child_frame is not None else self

        # Slider
        self.Slider = Slider(child_frame, self.__slider_callback,
                             bg=bg, line_color=line_color, active_line_color=active_line_color,
                             active_line_hover_color=active_line_hover_color,
                             slider_color=slider_color, slider_drag_color=slider_drag_color,
                             hide_slider=hide_slider, slider_visible=slider_visible,
                             slider_height=slider_height, slider_width=slider_width,
                             slider_type=slider_type, orientation=orientation,
                             height=canvas_height, width=canvas_width,
                             line_width=line_width)
        
        # Number Label
        label_bg = label_bg if label_bg is not None else bg
        label_fg = label_fg if label_fg is not None else text_fg
        label_hover_bg = label_hover_bg if label_hover_bg is not None else label_bg
        self.Label = NumberEditLabel(child_frame, self.__label_callback, min_value=min_value,
                                     max_value=max_value, step=step, default_value=default_value,
                                     reference_width=reference_width, max_len=max_len,
                                     drag_threshold=drag_threshold, draggable=label_draggable,
                                     editable=label_editable, justify=label_justify,
                                     bg=label_bg, fg=label_fg, hover_bg=label_hover_bg,
                                     error_color=label_error_color,
                                     entry_on_function=entry_on_function,
                                     entry_off_function=entry_off_function,
                                     font_name=font_name, font_size=font_size)
        
        # Static Text Label
        self.__popup_label = popup_label
        self.Text = Label(child_frame, text=label, bg=bg, fg=text_fg, font=(font_name, font_size))
        if self.__popup_label is not None:
            popup_font_name = popup_font_name if popup_font_name is not None else font_name
            popup_font_size = popup_font_size if popup_font_size is not None else font_size
            self.tool_tip = ToolTip(self.Text, bg=popup_bg, fg='#ffffff',
                                    font=(popup_font_name, popup_font_size))
            self.Text.bind('<Enter>', self.__text_hover_enter)
            self.Text.bind('<Leave>', self.__text_hover_leave)
        
        # Set Slider Value
        self.Slider.set(self.Label.get_perc())

    def __text_hover_enter(self, event=None):
        '''called when cursor enters the label'''
        if self.__popup_label is not None:
            self.tool_tip.fadein(0, self.__popup_label, event)

    def __text_hover_leave(self, event=None):
        '''called when cursor enters the label'''
        if self.__popup_label is not None:
            self.tool_tip.fadeout(1, event) # first argument is initial alpha
        
    def __slider_callback(self, perc:float):
        '''called when slider is moved by user
        :param perc: float between 0 and 1 - slider value
        '''
        self.Label.set_perc(perc)
        if self.__callback is not None:
            self.__callback(self.Label.get())

    def __label_callback(self, value:float):
        '''called when label value is changed by user'''
        self.Slider.set(self.Label.get_perc())
        if self.__callback is not None:
            self.__callback(value)

    def set_text(self, text:str):
        '''sets static text'''
        self.Text.config(text=text)

    def set_min_value(self, min_value):
        '''sets minimum value'''
        self.Label.set_min_value(min_value)
        self.Slider.set(self.Label.get_perc())

    def set_max_value(self, max_value):
        '''sets maximum value'''
        self.Label.set_max_value(max_value)
        self.Slider.set(self.Label.get_perc())

    def set(self, value):
        '''updates current value'''
        self.Label.set(value)
        self.Slider.set(self.Label.get_perc())

    def get(self):
        '''returns current value'''
        return self.Label.get()

class HorizontalSlider(LabelSlider):
    ''' Horizontal adaptation of LabelSlider
    
        Value label and static text label can be on either side of the slider

        There is the option for LabelSlider to be a single widget, or to grid
        the children directly in the parent frame. To grid children in the
        parent frame, specify the 'row' argument.
    '''
    def __init__(self, master:Frame, value_left=False, start_col=0, row=None,
                 slider_pady=0, **kwargs):
        '''
        Parameters
        ----------
            :param master: tk.Frame - parent frame
            :param value_left: bool - True to put value label on left side and text on right side
            :param start_col: int - start column (only if widgets are directly in master)
            :param row: int - row in master or None to put widgets in self
        '''
        child_frame = master if row is not None else None
        super().__init__(master, child_frame=child_frame, orientation='h', **kwargs)

        if row is not None: # to grid widgets in master
            if value_left:
                value_col, text_col = start_col, 2 + start_col
            else:
                value_col, text_col = 2 + start_col, start_col
            slider_col = 1 + start_col
            self.Slider.grid(row=row, column=slider_col, pady=slider_pady, sticky='ew')
            self.Label.grid(row=row, column=value_col, pady=slider_pady, sticky='nsew')
            self.Text.grid(row=row, column=text_col, pady=slider_pady, sticky='nsew')
        else: # pack in self
            if value_left:
                value_side, text_side = 'left', 'right'
            else:
                value_side, text_side = 'right', 'left'
            self.Label.pack(side=value_side)
            self.Text.pack(side=text_side)
            self.Slider.pack(side='left', fill='x', expand=True)

class VerticalSlider(LabelSlider):
    ''' Vertical adaptation of LabelSlider
    '''
    def __init__(self, master:Frame, **kwargs):
        super().__init__(master, orientation='v', label_justify='center', **kwargs)

        self.Label.pack(side='bottom', fill='x') # not sure if this should fill
        self.Slider.pack(side='bottom', fill='y', expand=True)
        if 'label' in kwargs.keys() and kwargs['label'] != '': # only pack Text if there is a label
            self.Text.pack(side='top')

class ScrollBar(Canvas):
    ''' ScrollBar for navigating scrollable widgets
    
        Only handles values between 0 and 1
        
        The current position is defined by two values that correspond to the
        visible start and end point of the linked scrollable widget
    '''
    def __init__(self, master:Frame, callback=None, default0=0, default1=1,
                 bg:str='#ffffff', slider_color:str='#000000',
                 slider_hover_color=None, slider_drag_color=None,
                 width=None, height=None, orientation='h',
                 disappear_when_filled=True, closeenough=1):
        '''
        Parameters
        ----------
            :param master: Frame - parent widget
            :param callback: function (float, float) - called when slider is moved
                                                     - takes two values between 0 and 1
            :param default0: float (between 0 and 1) - default lower bound
            :param default1: float (between 0 and 1) - default upper bound
            :param bg: str (hex code) - slider background color
            :param slider_color: str (hex code) - slider color when not hovering or dragging
            :param slider_hover_color: str (hex code) - slider color when hovering (if different from slider color)
            :param slider_drag_color: str (hex code) - slider color when dragging (if different from hover color)
            :param width: int (pixels) or None - canvas width (for vertical orientation)
            :param height: int (pixels) or None - canvas height (for horiontal orientation)
            :param orientation: Literal['h', 'v'] - horizontal or vertical orientation
            :param disappear_when_filled: bool - if True hides slider when no scrolling is possible
        '''
        assert orientation in ['h', 'v'], f'ScrollBar Error: Invalid orientation: {orientation}'
        assert default0 >= 0 and default0 <= 1, f'ScrollBar Error: Default0 out of range: {default0}'
        assert default1 >= 0 and default1 <= 1, f'ScrollBar Error: Default1 out of range: {default1}'

        super().__init__(master, bg=bg, width=width, height=height,
                         highlightthickness=0, closeenough=closeenough)

        self.__dragging = False
        self.__callback = callback
        self.__orientation = orientation
        self.__p0, self.__p1 = default0, default1
        self.__disappear_when_filled = disappear_when_filled
        self.__slider_color = slider_color
        self.__slider_hover_color = slider_hover_color if slider_hover_color is not None else self.__slider_color
        self.__slider_drag_color = slider_drag_color if slider_drag_color is not None else self.__slider_hover_color

        # {marker_id: {'id':canvas_id, 'value':value}}
        self.__markers = {}

        self.__slider_id = self.create_rectangle(0, 0, 0, 0, fill=self.__slider_color,
                                                 width=0, state='normal')
        
        # Event Bindings
        self.tag_bind(self.__slider_id, "<Button-1>", self.__click)
        self.tag_bind(self.__slider_id, "<ButtonRelease-1>", self.__release)
        self.tag_bind(self.__slider_id, "<B1-Motion>", self.__motion)
        self.tag_bind(self.__slider_id, "<Enter>", self.__hover_enter)
        self.tag_bind(self.__slider_id, "<Leave>", self.__hover_leave)
        
        self.bind('<Configure>', self.__set_coords)
        
        self.event_generate('<Configure>', when='tail')

    def __set_coords(self, event=None, filled_thresh=0.001):
        '''updates coordinates of lines and slider based on canvas size and current value'''
        # sets slider coordinates based on self.__p0 and self.__p1
        h, w = self.winfo_height(), self.winfo_width()
        if self.__orientation == 'h': # horizontal
            y0, y1 = 0, h
            x0, x1 = self.__p0 * w, self.__p1 * w
        elif self.__orientation == 'v': # vertical
            x0, x1 = 0, w
            y0, y1 = self.__p0 * h, self.__p1 * h
        self.coords(self.__slider_id, x0, y0, x1, y1)
        for marker in self.__markers.values():
            self.coords(marker['id'], *self.__get_marker_coords(marker['value'], h=h, w=w))
        if self.__disappear_when_filled and self.__p0 < filled_thresh and self.__p1 > 1 - filled_thresh:
            self.itemconfig(self.__slider_id, state='hidden')
        else:
            self.itemconfig(self.__slider_id, state='normal')

    def __get_marker_coords(self, value:float, h:int=None, w:int=None):
        '''returns (x0, y0, x1, y1) coordinates for the given marker value
            between 0 and 1
            Optionally pass canvas height, h and width, w to avoid recomputing
        '''
        if h is None:
            h = self.winfo_height()
        if w is None:
            w = self.winfo_width()
        if self.__orientation == 'h': # horizontal
            y0, y1 = 0, h
            x = value * w
            return x, y0, x, y1
        elif self.__orientation == 'v': # vertical
            x0, x1 = 0, w
            y = value * h
            return x0, y, x1, y
    
    # Cursor Callbacks
    def __hover_enter(self, event=None):
        '''cursor enters canvas'''
        self.itemconfig(self.__slider_id, fill=self.__slider_hover_color)

    def __hover_leave(self, event=None):
        '''cursor leaves canvas'''
        if not self.__dragging:
            self.itemconfig(self.__slider_id, fill=self.__slider_color)

    def __click(self, event):
        '''cursor clicks on canvas - move slider to click position'''
        self.__dragging = True
        self.itemconfig(self.__slider_id, fill=self.__slider_drag_color)
        x0, y0, _, _ = self.coords(self.__slider_id)
        self.__click_x = event.x - x0 # pixels from mouse to left edge of slider
        self.__click_y = event.y - y0 # pixels from mouse to top edge of slider

    def __release(self, event):
        '''cursor releases click - change slider color to normal'''
        self.__dragging = False
        self.itemconfig(self.__slider_id, fill=self.__slider_hover_color)
        x0, y0, x1, y1 = self.coords(self.__slider_id)
        if event.x < x0 or event.x > x1 or event.y < y0 or event.y > y1: # not hovering
            self.__hover_leave()

    def __motion(self, event):
        '''cursor drags slider'''
        x0, y0, x1, y1 = self.coords(self.__slider_id)
        w, h = self.winfo_width(), self.winfo_height()
        if self.__orientation == 'h':
            add = max(-x0, min(w - x1, (event.x - x0) - self.__click_x))
            x0 += add
            x1 += add
            self.__p0, self.__p1 = x0 / w, x1 / w
        elif self.__orientation == 'v':
            add = max(-y0, min(h - y1, (event.y - y0) - self.__click_y))
            y0 += add
            y1 += add
            self.__p0, self.__p1 = y0 / h, y1 / h
        self.coords(self.__slider_id, x0, y0, x1, y1)
        if self.__callback is not None:
            self.__callback(self.__p0, self.__p1)

    # Interface
    def set(self, p0:float, p1:float):
        '''sets slider value - must be between 0 and 1 (does not call callback function)'''
        assert p0 >= 0 and p0 <= 1, f'ScrollBar Error: set lower bound to invalid value: {p0}'
        assert p1 >= 0 and p1 <= 1, f'ScrollBar Error: set upper bound to invalid value: {p1}'
        assert p1 >= p0 ,f'ScrollBar Error: lower bound is greater than upper bound! lower: {p0}, upper: {p1}'
        self.__p0, self.__p1 = p0, p1
        self.__set_coords()

    def get(self):
        '''returns current slider value'''
        return self.__p0, self.__p1

    def add_marker(self, marker_id:str, value:float, color:str='#ffffff',
                   line_width:int=1):
        '''add marker line to ScrollBar'''
        assert value >= 0 and value <= 1, f'ScrollBar Error: Invalid marker value: {value}. Must be between 0 and 1'
        assert marker_id not in self.__markers.keys(), f'ScrollBar Error: marker_id already exists: {marker_id}'

        line_id = self.create_line(*self.__get_marker_coords(value), fill=color,
                                   width=line_width, state='disabled')
        
        self.__markers[marker_id] = {'id':line_id, 'value':value}

    def set_marker_value(self, marker_id:str, value:float):
        '''updates value of specified marker'''
        assert marker_id in self.__markers.keys(), f'ScrollBar Error: marker_id does not exist: {marker_id}'
        self.__markers[marker_id]['value'] = value
        self.coords(self.__markers[marker_id]['id'], *self.__get_marker_coords(value))

    def set_marker_color(self, marker_id:str, color:str):
        '''updates color of specified marker'''
        assert marker_id in self.__markers.keys(), f'ScrollBar Error: marker_id does not exist: {marker_id}'
        self.itemconfig(self.__markers[marker_id]['id'], fill=color)

    def delete_marker(self, marker_id:str):
        '''deletes specified marker'''
        assert marker_id in self.__markers.keys(), f'ScrollBar Error: trying to delete marker_id that does not exist: {marker_id}'
        self.delete(self.__markers[marker_id]['id'])
        del self.__markers[marker_id]

class HorizontalSliderGroup(Frame):
    ''' Group of horizontally oriented sliders connected by a single callback
        function
        
        The callback function is called whenever any of the slider values is
        changed by the user and is given 2 arguments: (slider_label, new_value)
    '''
    def __init__(self, master:Frame, parameters:list, callback, bg='#ffffff',
                 width=250, slider_pady=5, frame_padx=0, title_label=True,
                 **kwargs):
        '''Group of horizontal sliders aranged in a single column
        
        Parameters
        ----------
            :param master: tk.Frame - frame in which to put slider group
            :param parameters: list of dicts - each dict contains the keys:
                        label: str - slider name (given to callback function)
                        value: float - default value
                        min_value: float - minimum value
                        max_value: float - maximum value
                        step: float - slider increment
                        description: str - info displayed when mouse hovers on slider (optional)
            :param callback: function (parameter_name, value) - called whenever a slider is adjusted
            :param bg: str (hex code) - background color
            :param width: int - width of each slider in pixels
            :param slider_pady: int - y pad inside sliders
            :param frame_padx: int - x pad inside frame
            :param title_label: bool - if True display 'fade_in' as 'Fade In'
            **kwargs passed to each slider
        '''
        Frame.__init__(self, master, bg=bg, padx=frame_padx)
        self.grid_columnconfigure(0, weight=0) # text label
        self.grid_columnconfigure(1, weight=1) # slider
        self.grid_columnconfigure(2, weight=0) # value label
        self.sliders: list[HorizontalSlider] = []
        self.labels = [d['label'] for d in parameters]
        for i, param in enumerate(parameters):
            label = param['label'].replace('_', ' ').title() if title_label else param['label']
            popup_label = param['description'] if 'description' in param.keys() else None
            S = HorizontalSlider(self, start_col=0, row=i, bg=bg, canvas_width=width,
                                 callback=lambda x, l=param['label']: callback(l, x),
                                 min_value=param['min_value'], max_value=param['max_value'],
                                 step=param['step'], default_value=param['value'],
                                 label=label, popup_label=popup_label,
                                 slider_pady=slider_pady, **kwargs)
            self.sliders.append(S)

    def set(self, d:dict):
        '''sets values of sliders - input dictionary MUST contain key and value for each slider'''
        for label, slider in zip(self.labels, self.sliders):
            slider.set(d[label])

    def get(self) -> dict:
        '''returns dictionary with current value of each slider'''
        return {label:slider.get() for label, slider in zip(self.labels, self.sliders)}

class VerticalSliderGroup(Frame):
    ''' Group of vertically oriented sliders connected by a single callback
        function
        
        The callback function is called whenever any of the slider values is
        changed by the user and is given 2 arguments: (slider_label, new_value)
    '''
    def __init__(self, master, parameters:list, callback, bg:str, rows:int,
                 columns:int, height=250, slider_pady=5, slider_padx=5,
                 title_label=True, **kwargs):
        '''Group of vertical sliders aranged in rows and columns
        
        Parameters
        ----------
            :param master: tk.Frame - frame in which to put slider group
            :param parameters: list of dicts - each dict contains the keys:
                        label: str - slider name (given to callback function)
                        value: float - default value
                        min_value: float - minimum value
                        max_value: float - maximum value
                        step: float - slider increment
                        description: str - info displayed when mouse hovers on slider
            :param callback: function (parameter_name, value) - called whenever a slider is adjusted
            :param bg: str (hex code) - background color
            :param rows: int - rows of sliders - filling begins rowwise at top left
            :param columns: int - columns of sliders - filling begins rowwise at top left
            :param height: int - height of each slider in pixels
            :param slider_pady: int - y pad between sliders
            :param slider_padx: int - x pad between sliders
            :param title_label: bool - if True display 'fade_in' as 'Fade In'
            **kwargs passed to each slider
        '''
        Frame.__init__(self, master, bg=bg)
        for i in range(rows):
            self.grid_rowconfigure(i, weight=1)
        for i in range(columns):
            self.grid_columnconfigure(i, weight=1)
        self.sliders: list[VerticalSlider] = []
        self.labels = [d['label'] for d in parameters]
        for i, param in enumerate(parameters):
            label = param['label'].replace('_', ' ').title() if title_label else param['label']
            popup_label = param['description'] if 'description' in param.keys() else None
            S = VerticalSlider(self, callback=lambda x, l=param['label']: callback(l, x),
                               min_value=param['min_value'], max_value=param['max_value'],
                               step=param['step'], default_value=param['value'],
                               label=label, popup_label=popup_label, bg=bg,
                               canvas_height=height, **kwargs)
            S.grid(row=i // columns, column=i % columns, sticky='nsew',
                   padx=slider_padx, pady=slider_pady)
            self.sliders.append(S)

    def set(self, d:dict):
        '''sets values of sliders - input dictionary MUST contain key and value for each slider'''
        for label, slider in zip(self.labels, self.sliders):
            slider.set(d[label])

    def get(self) -> dict:
        '''returns dictionary with current value of each slider'''
        return {label:slider.get() for label, slider in zip(self.labels, self.sliders)}
   
class PlotScrollBar(Canvas):
    ''' Horizontally oriented plot slider that handles integer values between
        a specified minimum and maximum value
        
        Callback command is called whenever the slider is moved by the user

        Values are displayed to the user in seconds (formatted m:ss), but on the
        backend, values are given in frames. This is ideal if you are using the
        slider to control video playback
    '''
    def __init__(self, master, command, label, frames=100, min_frame=0, start_frame=1,
                 frame_rate=29.97, height=72, padx=0.04, active_color='#e8ff00',
                 inactive_color='#888888', hover_color='#ffffff',
                 active_fill_color=brighten('#00ff00', -0.75), bg='#000000',
                 show_fill=False, active_fill=False, confine_to_active_region=False,
                 fill_text='', active_x0=0, active_x1=0, mouse_wheel_steps=1,
                 active_fill_callback=None, font_name='Segoe UI', label_font_size=10,
                 tick_font_size=9, active=True):
        '''        
        Parameters
        ----------
            :param command: function (int) - called whenever slider value is changed
            :param label: str or None - x axis label display beneath slider
            :param frames: int - total number of steps in slider
            :param start_frame: int - default slider position
            :param height: int - window height in pixels
            :param mouse_wheel_steps: int - increment for each mousewheel scroll event
            :param show_fill: bool - if True, display fill of active range underneath slider - must be True to use confine_to_active_region
            :param active_fill: bool - if True and show_fill is True, fill of active range can be dragged
            :param confine_to_active_region: bool - if True, user will not be allowed to move scrollbar outside of active region
            :param fill_text: str - text displayed in fill of active range
            :param (active_x0 active_x1) : (int, int) - start and end of active range
            :param active_fill_callback: 2 argument function (active_x0, active_x1) - called whenever active region is changed
            :param active: bool - if False, scrollbar will be unresponsive to user interactions - for toggling
        '''
        Canvas.__init__(self, master, bg=bg, height=height, highlightthickness=0)
        self.command = command # function acception 1 numeric argument
        self.active_fill_callback = active_fill_callback
        self.label = label # can be none to display no label
        self.mouse_wheel_steps = mouse_wheel_steps
        self.height, self.padx = height, padx
        self.active_color, self.inactive_color, self.hover_color = active_color, inactive_color, hover_color
        self.font_name, self.label_font_size, self.tick_font_size = font_name, label_font_size, tick_font_size
        self.show_fill, self.active_fill = show_fill, active_fill
        self.confine_to_active_region = confine_to_active_region
        self.hovering, self.dragging = False, False
        self.frame_rate = frame_rate
        self.active_x0, self.active_x1 = active_x0, active_x1 # frames
        self.current_frame, self.min_frame, self.max_frame = start_frame, min_frame, frames
        self.min_seconds, self.max_seconds = self.min_frame / frame_rate, self.max_frame / frame_rate
        self.width = None
        self.xmin, self.xmax = 0, 1 # only needed before Canvas has been rendered - '<Configure>'
        if self.label:
            self.label_line_height, self.tick_height = height * 0.5, height * 0.59
        else:
            self.label_line_height, self.tick_height = height * 0.65, height * 0.76
        self.divisions = np.array([1/2**i for i in range(4, 0, -1)] + [1, 2, 5, 10, 25, 50, 100]) # smallest is 0.0625 (1/16)

        if self.show_fill:
            self.Fill = CanvasEditFill(self, 'sb_fill', 0, 0, 0, self.label_line_height, line_width=2, bg=active_fill_color,
                                       main_text_hover_justify='center', main_text=fill_text, selectable=False, hoverable=True,
                                       box_draggable='horizontal', left_draggable=active_fill, right_draggable=active_fill,
                                       left_drag_function=self.active_drag0, right_drag_function=self.active_drag1,
                                       box_drag_function=self.active_drag, brighten_fact=0, raise_on_click=False,
                                       active=active and self.active_fill)
        self.Line = CanvasEditLine(self, 0, 0, self.label_line_height, width=6,
                                   bg=self.inactive_color, drag_color=self.active_color,
                                   hover_color=self.hover_color, selectable=False,
                                   hoverable=True, draggable=True, deletable=False,
                                   show_drag_color=True, drag_function=self.main_drag,
                                   active=active)

        self.bind('<Configure>', self.frame_width)
        self.bind('<MouseWheel>', self.mouse_wheel_scroll)
        self.event_generate('<Configure>', when='tail')

    def frame_width(self, event):
        '''called whenever window is resized'''
        self.width = event.width
        self.xmin = self.width * self.padx
        self.xmax = self.width * (1 - self.padx)
        self.draw()

    def get_status(self):
        '''returns current_frame, min_frame, and max_frame'''
        return self.current_frame, self.min_frame, self.max_frame
    
    def get_current_frame(self):
        '''returns current frame of ScrollBar'''
        return self.current_frame

    def set_frame_num(self, max_frame:int, frame_rate:float, min_frame:int=0):
        '''update the frame bounds and frame rate'''
        self.frame_rate = frame_rate
        self.min_frame, self.max_frame = min_frame, max_frame
        self.min_seconds, self.max_seconds = self.min_frame / self.frame_rate, self.max_frame / self.frame_rate
        self.update_active_region()
        self.draw()

    def set_frame(self, frame:int):
        '''set the slider position'''
        self.current_frame = frame
        self.update_line_x()

    def set_active(self):
        '''sets state to active so that scrollbar will be responsive to user interactions'''
        self.Line.set_active()
        if self.show_fill and self.active_fill: # if fill is set to be uninteractable, Fill will remain inactive
            self.Fill.set_active()

    def set_inactive(self):
        '''sets state to inactive so that scrollbar will be unresponsive to user interactions'''
        self.Line.set_inactive()
        if self.show_fill:
            self.Fill.set_inactive()

    def update_active_fill(self, start_frame:int, end_frame:int):
        '''configure active region'''
        # could throw error if new active region extends beyond scrollbar limits (self.min_frame to self.max_frame)
        if not self.show_fill:
            # raise error
            return
        self.active_x0, self.active_x1 = start_frame, end_frame
        self.update_fill_x()
            
    def get_frame_x(self, frame:int) -> float:
        '''returns the x coordinate corresponding to the given frame'''
        if self.max_frame - self.min_frame == 0:
            return self.xmin
        return self.xmin + (self.xmax - self.xmin) * (frame - self.min_frame) / (self.max_frame - self.min_frame)
    
    def x_coord_to_frame(self, x:float) -> int:
        '''converts x coordinate to frame - returns int'''
        x_perc = (x - self.xmin) / (self.xmax - self.xmin)
        frame = int(self.min_frame + (self.max_frame - self.min_frame) * x_perc)
        return min(self.max_frame, max(self.min_frame + 1, frame))
            
    def update_line_x(self):
        '''updates line based on self.current_frame and moves active region if necessary'''
        self.Line.set_x(self.get_frame_x(self.current_frame))
        self.update_active_region()

    def update_fill_x(self, callback=True):
        '''updates fill position based on self.active_x0 and self.active_x1'''
        if self.show_fill and self.width: # dont actually do changes if scrollbar has not been drawn yet
            self.Fill.set_x_coords(self.get_frame_x(self.active_x0), self.get_frame_x(self.active_x1))
        if callback and self.active_fill_callback:
            self.active_fill_callback(self.active_x0, self.active_x1)

    def squeeze_active_bounds(self, active_x0:int, active_x1:int):
        '''
        Purpose
        -------
            adjusts active_x0 and active_x1 if necessary based on the following conditions
                active_x0 must be >= self.min_frame
                active_x0 must be < self.current_frame
                active_x1 must be <= self.max_frame
                active_x1 must be > self.current_frame
            distance between active_x0 and active_x1 will remain the same
                
        Parameters
        ----------
            active_x0 : int - left edge of active region before adjustment
            active_x1 : int - right edge of active region before adjustment
            
        Returns
        -------
            active_x0 : int - left edge of active region after adjustment
            active_x1 : int - right edge of active region after adjustment
        '''
        width = max(2, min(self.max_frame - self.min_frame, active_x1 - active_x0))
        x0, x1 = active_x0, active_x0 + width
        if x0 < self.min_frame:
            add = self.min_frame - x0
            x0 += add
            x1 += add
        elif x0 > self.current_frame - 1:
            sub = x0 - (self.current_frame - 1)
            x0 -= sub
            x1 -= sub
        elif x1 < self.current_frame + 1:
            add = self.current_frame + 1 - x1
            x0 += add
            x1 += add
        elif x1 > self.max_frame:
            sub = x1 - self.max_frame
            x0 -= sub
            x1 -= sub
        return x0, x1

    def update_active_region(self):
        '''updates active region if it does not include self.current_frame
        to be called when main line is moved
        other rules may be set to stop the main line from being at the very edge of active region
        '''
        if self.show_fill:
            if self.active_x0 in range(self.min_frame, self.current_frame) and self.active_x1 in range(self.current_frame + 1, self.max_frame + 1):
                return # active bounds are both within allowed range
            self.active_x0, self.active_x1 = self.squeeze_active_bounds(self.active_x0, self.active_x1)
            self.update_fill_x()

    def main_drag(self, current_x:float, cursor_x:float):
        '''called whenever main line is dragged
        move main line and squeeze inside bounds if necessary
        
        Parameters
        ----------
            current_x : float - position of line before drag
            cursor_x : float - cursor position - potential new line position
        
        Returns
        -------
            new_x : float - new position for line
        '''
        frame = self.x_coord_to_frame(cursor_x)
        if self.confine_to_active_region: # do not allow user to move scrollbar outside of active region
            frame = min(self.active_x1 - 1, max(self.active_x0 + 1, frame))
        if frame != self.current_frame:
            self.current_frame = frame
            self.command(self.current_frame - 1)
            self.update_active_region()
        return self.get_frame_x(self.current_frame) # to update line position
    
    def active_drag(self, x0:float, x1:float, cursor_x:float, new_x0:float, new_x1:float):
        '''called whenever the active fill is dragged
        moves active region - cannot exclude main line (current frame)
        
        Parameters
        ----------
            x0 : float - position of left edge of fill before drag
            x1 : float - position of right edge of fill before drag
            cursor_x : float - cursor position
            new_x0 : float - potential new left edge based on cursor movement
            new_x1 : float - potential new right edge based on cursor movement
        
        Returns
        -------
            new_x0 : float - new position of left edge of fill
            new_x1 : float - new position of right edge of fill
        '''
        active_x0, active_x1 = self.x_coord_to_frame(new_x0), self.x_coord_to_frame(new_x1)
        active_x0, active_x1 = self.squeeze_active_bounds(active_x0, active_x1)
        if active_x0 != self.active_x0 and active_x1 != self.active_x1: # width of active region must not change
            self.active_x0, self.active_x1 = active_x0, active_x1
            if self.active_fill_callback:
                self.active_fill_callback(self.active_x0, self.active_x1)
        return self.get_frame_x(self.active_x0), self.get_frame_x(self.active_x1)

    def active_drag0(self, current_x0:float, current_x1:float, cursor_x:float):
        '''called whenever the left active line is dragged
        moves active region - cannot exclude main line (current frame)
        
        Parameters
        ----------
            current_x0 : float - position of left edge before drag
            current_x1 : float - position of right edge before drag
            cursor_x : float - cursor position - potential new left edge position
        
        Returns
        -------
            new_x : float - new position for line
        '''
        frame = min(self.current_frame - 1, max(self.min_frame, self.x_coord_to_frame(cursor_x))) # cannot be >= than current frame
        if frame != self.active_x0:
            self.active_x0 = frame
            if self.active_fill_callback:
                self.active_fill_callback(self.active_x0, self.active_x1)
        return self.get_frame_x(self.active_x0) # to update fill position
    
    def active_drag1(self, current_x0:float, current_x1:float, cursor_x:float):
        '''called whenever the right active line is dragged
        moves active region - cannot exclude main line (current frame)
        
        Parameters
        ----------
            current_x0 : float - position of left edge before drag
            current_x1 : float - position of right edge before drag
            cursor_x : float - cursor position - potential new right edge position
        
        Returns
        -------
            new_x : float - new position for line
        '''
        frame = min(self.max_frame, max(self.current_frame + 1, self.x_coord_to_frame(cursor_x))) # cannot be <= than current frame
        if frame != self.active_x1:
            self.active_x1 = frame
            if self.active_fill_callback:
                self.active_fill_callback(self.active_x0, self.active_x1)
        return self.get_frame_x(self.active_x1) # to update fill position

    def increment_frame(self, increment, loop=False, callback=True):
        '''returns False if the end has been reached or no effect is loaded, otherwise True
        calls callback with new frame'''
        if loop and self.current_frame == self.min_frame + 1 and increment == -1: # reached the beginning when going backward with loop
            self.current_frame = self.max_frame - 1
        if self.current_frame == self.max_frame and increment >= 0: # reached end
            if loop: # loop back to start
                self.current_frame = self.min_frame + 1
                self.update_line_x() # based on self.current_frame
                if callback:
                    self.command(self.current_frame - 1)
                return True
            else: # end
                return False
        self.current_frame = int(min(self.max_frame, max(self.min_frame + 1, self.current_frame + increment)))
        self.update_line_x()
        if callback:
            self.command(self.current_frame - 1)
        return True

    def mouse_wheel_scroll(self, event):
        '''event.delta / 120 is float - number of scroll steps - positive for up, negative for down'''
        frame = min(self.max_frame, max(self.min_frame + 1, int(self.current_frame + event.delta / 120 * self.mouse_wheel_steps)))
        if self.confine_to_active_region: # do not allow user to move scrollbar outside of active region
            frame = min(self.active_x1 - 1, max(self.active_x0 + 1, frame))
        self.set_frame(frame)
        self.command(self.current_frame - 1)

    def remove(self):
        '''removes all labels and ticks from the x axis'''
        # deletes everything except line and active fill
        for i in self.find_all():
            if i != self.Line.id and (not self.show_fill or i != self.Fill.box_id):
                self.delete(i)

    def draw(self, max_ticks=15, tick_x_buffer=0.01):
        '''draws label line and ticks/labels on canvas
        updates position of main scroll line and active region fill'''
        if not self.width:
            # width will not be set for sliders on subsequent pages that have not be loaded yet
            return None
        self.remove()
        self.update_line_x()
        self.update_fill_x(callback=False)
        self.create_line(self.xmin, self.label_line_height, self.xmax,
                         self.label_line_height, fill='#ffffff', width=1)
        if self.label != None:
            self.create_text(self.width / 2, self.height, text=self.label,
                             fill='#ffffff', font=(self.font_name, self.label_font_size),
                             anchor='s')
        
        # draw ticks and labels
        increment = self.divisions[self.divisions > (self.max_seconds - self.min_seconds) / max_ticks].min()
        ticks = np.arange(int(self.min_seconds / increment), int(self.max_seconds / increment) + 1) * increment # in seconds
        for sec, label in zip(ticks, [seconds_text(t) for t in ticks]):
            x = self.xmin + (self.xmax - self.xmin) * (sec - self.min_seconds) / (self.max_seconds - self.min_seconds)
            if x + (self.xmax - self.xmin) * tick_x_buffer >= self.xmin and x - (self.xmax - self.xmin) * tick_x_buffer <= self.xmax:
                self.create_line(x, self.label_line_height, x, self.tick_height,
                                 fill='#ffffff', width=1)
                self.create_text(x, self.tick_height, text=label, fill='#ffffff',
                                 font=(self.font_name, self.tick_font_size), anchor='n')

class DoubleScrollBar(Frame):
    ''' Extension of PlotScrollBar that combines to PlotScrollBars to give user
        more precise control
        
        The top scrollbar displays an adjustable active region, which defines
        the bounds of the bottom scrollbar. User can use the bottom scrollbar
        to seek within the active region.

        Just like PlotScrollBar, there is a single callback command that is
        called whenever the slider value is changed, either by the top or bottom
        PlotScrollBar
    '''
    def __init__(self, master, command, label, frames=100, min_frame=0, start_frame=1,
                 frame_rate=29.97, main_height=60, secondary_height=60, padx=0.04,
                 active_color='#e8ff00', inactive_color='#888888',
                 hover_color='#ffffff', bg='#000000', active_fill_color='#005500',
                 fill_text='', mouse_wheel_steps=1, secondary_width_perc=0.2,
                 font_name='Segoe UI', label_font_size=10, tick_font_size=9,
                 confine_to_active_region=False, active_fill=True, active=True):
        '''Double version of PlotScrollBar - main scrollbar on top and secondary scrollbar beneath for precise seeking
        top scrollbar has interactable fill to control the bounds of bottom scrollbar
        
        Parameters
        ----------
            :param command: 1 argument function (int) - called whenever slider value is changed
            :param label: str or None - x axis label display beneath bottom slider
            :param frames: int - total number of steps in slider
            :param start_frame: int - default slider position
            :param height: int - window height in pixels
            :param mouse_wheel_steps: int - increment for each mousewheel scroll event
            :param fill_text: str - text displayed in fill of active range
            :param secondary_width_perc: float between 0 and 1 - range of secondary scrollbar as a fraction of main scrollbar range
            :param confine_to_active_region: bool - if True, user will not be allowed to move scrollbar outside of active region
            :param active_fill: bool - if True, active region of MainScrollBar can be adjusted by user
            :param active: bool - if False, scrollbar will be unresponsive to user interactions - for toggling
        '''
        Frame.__init__(self, master, bg=bg)
        self.command = command
        self.secondary_width_perc = secondary_width_perc
        self.frame_rate = frame_rate

        active_x0, active_x1 = self.get_active_bounds(start_frame, min_frame, frames, self.secondary_width_perc)   
        fill_callback = lambda f0, f1: self.SecondaryScrollBar.set_frame_num(f1, self.frame_rate, min_frame=f0 + 1)     
        self.MainScrollBar = PlotScrollBar(self, self.__main_command, None,
                                           frames=frames, min_frame=min_frame,
                                           start_frame=start_frame, height=main_height,
                                           padx=padx, active_color=active_color,
                                           inactive_color=inactive_color,
                                           hover_color=hover_color,
                                           active_fill_color=active_fill_color,
                                           bg=bg, show_fill=True, active_fill=active_fill,
                                           fill_text=fill_text, active_x0=active_x0,
                                           active_x1=active_x1, mouse_wheel_steps=mouse_wheel_steps,
                                           font_name=font_name, label_font_size=label_font_size,
                                           tick_font_size=tick_font_size, active=active,
                                           confine_to_active_region=confine_to_active_region,
                                           active_fill_callback=fill_callback,
                                           frame_rate=self.frame_rate)
        self.SecondaryScrollBar = PlotScrollBar(self, self.__secondary_command,
                                                label, frames=self.MainScrollBar.active_x1,
                                                min_frame=self.MainScrollBar.active_x0,
                                                start_frame=start_frame,
                                                height=secondary_height, padx=padx,
                                                active_color=active_color, inactive_color=inactive_color,
                                                hover_color=hover_color, active_fill_color=active_fill_color, bg=bg,
                                                show_fill=False, active_fill=False,
                                                mouse_wheel_steps=mouse_wheel_steps,
                                                font_name=font_name, label_font_size=label_font_size,
                                                tick_font_size=tick_font_size, active=active,
                                                frame_rate=self.frame_rate)
        self.MainScrollBar.pack(side='top', fill='x')
        self.SecondaryScrollBar.pack(side='top', fill='x')

    def get_active_bounds(self, start_frame:int, min_frame:int, max_frame:int, width_perc:float):
        '''returns bounds of active region based
        
        Parameters
        ----------
            :param start_frame: int - current frame of slider
            :param min_frame: int - minimum frame - left edge of scrollbar
            :param max_frame: int - maximum frame - right edge of scrollbar
            :param width_perc: float between 0 and 1 - range of secondary scrollbar as a fraction of main scrollbar range
            
        Returns
        -------
            :return x0: int - left edge of active region
            :return x1: int - right edge of active region
        '''
        pad = (max_frame - min_frame) * width_perc / 2
        x0, x1 = start_frame - pad, start_frame + pad
        if x0 < min_frame:
            x0, x1 = min_frame, x1 + min_frame - x0
        if x1 > max_frame:
            x0, x1 = x0 - (x1 - max_frame), max_frame
        return int(x0), int(x1)
        
    def __main_command(self, frame:int):
        '''called internally when main scrollbar is moved by user'''
        self.SecondaryScrollBar.set_frame(frame)
        self.command(frame)

    def __secondary_command(self, frame:int):
        '''called internally when secondary scrollbar is moved by user'''
        self.MainScrollBar.set_frame(frame)
        self.command(frame)

    def set_frame_num(self, max_frame:int, frame_rate:float, min_frame:int=0):
        '''updates the min/max frame and frame rate'''
        self.MainScrollBar.set_frame_num(max_frame, frame_rate, min_frame=min_frame)
        self.MainScrollBar.update_active_fill(*self.get_active_bounds(*self.MainScrollBar.get_status(), self.secondary_width_perc))

    def set_frame(self, frame):
        '''set current frame - pushes to both scrollbars'''
        # this will update active region if necessary and propogate to SecondaryScrollBar
        self.MainScrollBar.set_frame(frame)
        self.SecondaryScrollBar.set_frame(frame)

    def set_active(self):
        '''sets state to active so that scrollbar will be responsive to user interactions'''
        self.MainScrollBar.set_active()
        self.SecondaryScrollBar.set_active()

    def set_inactive(self):
        '''sets state to inactive so that scrollbar will be unresponsive to user interactions'''
        self.MainScrollBar.set_inactive()
        self.SecondaryScrollBar.set_inactive()

    def get_current_frame(self):
        '''returns current frame of ScrollBar'''
        return self.MainScrollBar.get_current_frame()

    def update_active_fill(self, start_frame:int, end_frame:int):
        '''
        update region that can be scrolled by the bottom scrollbar
        this is the region displayed as active in the top scrollbar
        '''
        self.MainScrollBar.update_active_fill(start_frame, end_frame)

    def increment_frame(self, increment, loop=False, callback=True):
        '''returns False if the end has been reached or no effect is loaded, otherwise True
        calls callback with new frame'''
        current_frame, min_frame, max_frame = self.MainScrollBar.get_status()
        if current_frame + increment <= min_frame and increment < 0 and loop: # reached the beginning when going backward with loop
            new_frame = max_frame - 1
        elif current_frame + increment >= max_frame and increment > 0 and loop: # reached end with loop (going forward)
            new_frame = min_frame + 1
        else:
            new_frame = current_frame + increment

        set_frame = int(min(max_frame, max(min_frame + 1, new_frame)))
        self.set_frame(set_frame)
        if callback:
            self.command(set_frame - 1)
        return new_frame in range(min_frame, max_frame + 1)




from tkinter import Widget, Event

from .buttons import IconButton
from .tool_frame import ToolFrame
from .icons import icons


class LabelDropdown(IconButton):
    ''' Combines IconButton and ToolFrame to show dropdown menu with
        selectable labels when button is clicked
    '''
    def __init__(self, master:Widget, parameters:list, callback=None, icon=None,
                 popup_label=None, bg:str='#000000', inactive_fg:str='#888888',
                 hover_fg:str='#ffffff', label_bg:str=None, label_active_bg:str=None,
                 label_fg:str=None, label_active_fg:str=None):
        '''
        Parameters
        ----------
            :param master: tk.Widget - parent widget (for button)
            :param parameters: list[dict] - each dict has keys: ['label', 'icon']
            :param callback: function(index:int, status:bool) - called when a label is clicked
            :param icon: str (path to png), np.array or None (for default menu icon)
            :param popup_label: str - displayed when cursor hovers on button
        '''
        # Configure label colors
        label_bg = label_bg if label_bg is not None else bg
        label_active_bg = label_active_bg if label_active_bg is not None else label_bg
        label_fg = label_fg if label_fg is not None else inactive_fg
        label_active_fg = label_active_fg if label_active_fg is not None else hover_fg

        icon = icon if icon is not None else icons['menu']
        super().__init__(master, icon, self.__button_click, bar_height=0,
                         selectable=False, inactive_bg=bg, inactive_fg=inactive_fg,
                         inactive_hover_fg=hover_fg, popup_label=popup_label)
        
        self.__ToolFrame = ToolFrame(self, parameters, callback, bg=label_bg,
                                     active_bg=label_active_bg, fg=label_fg,
                                     active_fg=label_active_fg)
        
        self.bind('<Leave>', self.__button_leave, add='+')
        self.__ToolFrame.bindall('<Leave>', self.__popup_leave, add='+')

    # Event Callbacks
    def __button_leave(self, event:Event):
        '''called when cursor leaves button'''
        self.after(200, lambda: self.__leave(event))
    
    def __popup_leave(self, event:Event):
        '''called when cursor leaves button'''
        self.after(200, lambda: self.__leave(event))

    def __leave(self, event:Event):
        '''called a few miliseconds after the cursor leaves the button or popup'''
        if not self.is_hovering() and not self.__ToolFrame.contains_point(event.x_root, event.y_root):
            self.__ToolFrame.hide()

    def __button_click(self, event=None):
        '''user clicks on menu button to display popup'''
        self.__ToolFrame.show()

    # Pass to ToolFrame
    def select_all(self):
        '''selects all labels'''
        self.__ToolFrame.select_all()
    
    def deselect_all(self):
        '''deselects all labels'''
        self.__ToolFrame.deselect_all()

    def set(self, d:dict):
        '''sets selection status of labels
        :param d: dict - label(str):status(bool) for each item
        '''
        self.__ToolFrame.set(d)

    def get(self) -> dict:
        '''returns dictionary with label(str):status(bool) for each item'''
        return self.__ToolFrame.get()



from tkinter import Label, Frame


class CollapseFrame(Frame):
    ''' Extension of tk.Frame that includes a header that is always visible,
        and a body frame that can be shown/hidden by clicking on the header
        or with an external command.

        Contains the attributes 'header' and 'frame' which are both tk.Frame
        objects refering to the header and body frames respectively.

        DO NOT put anything directly in the CollapseFrame. Only user the header
        and body frames.

        Warning: if CollapseFrame is initially close, and not collapsable,
        there will be no way for user to access the content of the body frame.
    '''
    def __init__(self, master:Frame, open_callback=None, close_callback=None,
                 label=None, bg='#ffffff', border_bg=None,
                 inactive_bg=None, active_bg=None, inactive_hover_bg=None,
                 active_hover_bg=None, inactive_fg='#000000', active_fg=None,
                 inactive_hover_fg=None, active_hover_fg=None,
                 label_side='left', font_name='Segoe UI', font_size=10,
                 label_padx=0, label_pady=0, body_padx=0, body_pady=0,
                 active=True, collapsable=True, **kwargs):
        '''active refers to the state when the body frame is visible
        Parameters
        ----------
            :param master: tk.Frame - parent widget
            :param open_callback: function () - called when body frame is made visible
            :param close_callback: function () - called when body frame is hidden
            :param label: str or None - header text
            :param bg: str (hex code) - background color
            :param border_bg: str (hex code) - color of base frame (border around body frame)
            :param inactive_bg: str (hex code) - inactive header background color (if different from bg)
            :param active_bg: str (hex code) - active header background color (if different from inactive_bg)
            :param inactive_hover_bg: str (hex code) - inactive hover background color (if different from inactive_bg)
            :param active_hover_bg: str (hex code) - active hover background color (if different from active_bg)
            :param inactive_fg: str (hex code) - inactive label color
            :param active_fg: str (hex code) - active label color
            :param inactive_hover_fg: str (hex code) - inactive hover label color (if different from inactive_fg)
            :param active_hover_fg: str (hex code) - active hover label color (if different from active_fg)
            :param label_side: str Literal['left', 'right', 'top'] - label justification
            :param font_name: str - label font name
            :param font_size: str - label font size
            :param label_padx: int - x padding of label within header frame
            :param label_pady: int - y padding of label within header frame
            :param body_padx: int - x padding around body frame
            :param body_pady: int - y padding around body frame
            :param active: bool - initial visibility of body frame
            :param collapsable: bool - if True, CollapseFrame is static
        '''
        super_bg = border_bg if border_bg is not None else bg
        super().__init__(master, bg=super_bg, **kwargs)
        self.__open_callback = open_callback
        self.__close_callback = close_callback
        self.__body_padx, self.__body_pady = body_padx, body_pady
        inactive_bg = inactive_bg if inactive_bg is not None else bg
        active_bg = active_bg if active_bg is not None else inactive_bg
        inactive_hover_bg = inactive_hover_bg if inactive_hover_bg is not None else inactive_bg
        active_hover_bg = active_hover_bg if active_hover_bg is not None else active_bg
        self.__bg_colors = [[inactive_bg, active_bg], [inactive_hover_bg, active_hover_bg]]
        active_fg = active_fg if active_fg is not None else inactive_fg
        inactive_hover_fg = inactive_hover_fg if inactive_hover_fg is not None else inactive_fg
        active_hover_fg = active_hover_fg if active_hover_fg is not None else active_fg
        self.__fg_colors = [[inactive_fg, active_fg], [inactive_hover_fg, active_hover_fg]]
        self.__draw_label = label is not None
        self.__collapsable = collapsable
        self.__active, self.__hovering = active, False

        # Header Frame
        self.header = Frame(self)
        self.header.pack(side='top', fill='x')
        if self.__collapsable:
            self.header.bind('<Enter>', self.hover_enter)
            self.header.bind('<Leave>', self.hover_leave)
            self.header.bind('<Button-1>', self.header_click)
        if self.__draw_label:
            self.label = Label(self.header, text=label, font=(font_name, font_size))
            self.label.pack(side=label_side, padx=label_padx, pady=label_pady)
            if self.__collapsable:
                self.label.bind('<Enter>', self.hover_enter)
                self.label.bind('<Leave>', self.hover_leave)
                self.label.bind('<Button-1>', self.header_click)

        # Body Frame
        self.frame = Frame(self, bg=bg)
        if self.__active:
            self.show(callback=False)

        self.__color_config()

    def __color_config(self):
        '''updates header color based on active/hover status'''
        bg = self.__bg_colors[self.__hovering][self.__active]
        fg = self.__fg_colors[self.__hovering][self.__active]
        self.header.config(bg=bg)
        if self.__draw_label:
            self.label.config(bg=bg, fg=fg)

    def set_bg(self, color:str, hovering=False, active=False):
        '''updates background color for given hover/select status'''
        self.__bg_colors[hovering][active] = color

    def set_fg(self, color:str, hovering=False, active=False):
        '''updates foreground color for given hover/select status'''
        self.__fg_colors[hovering][active] = color

    def hover_enter(self, event=None):
        '''called when mouse enters header'''
        if not self.__collapsable:
            print('CollapseFrame Error: tried to hover enter when CollapseFrame is not collapsable')
            return
        self.__hovering = True
        self.__color_config()

    def hover_leave(self, event=None):
        '''called when mouse leaves header'''
        if not self.__collapsable:
            print('CollapseFrame Error: tried to hover leave when CollapseFrame is not collapsable')
            return
        self.__hovering = False
        self.__color_config()

    def header_click(self, event=None):
        '''called when mouse clicks on header - toggles active status'''
        if not self.__collapsable:
            print('CollapseFrame Error: tried to toggle body frame when CollapseFrame is not collapsable')
            return
        if self.__active:
            self.hide(callback=True)
        else:
            self.show(callback=True)

    def show(self, callback=True):
        '''makes body frame visible'''
        self.__active = True
        self.__color_config()
        self.frame.pack(fill='both', expand=True,
                        padx=self.__body_padx, pady=self.__body_pady)
        if callback and self.__open_callback is not None:
            self.__open_callback()

    def hide(self, callback=True):
        '''hides body frame'''
        self.__active = False
        self.__color_config()
        self.frame.pack_forget()
        if callback and self.__close_callback is not None:
            self.__close_callback()




from tkinter import Frame


class AspectFrame(Frame):
    ''' Dynamic Frame that uses flexible border padding to maintain a
        consistent aspect ratio

        All content should be put in AspectFrame.frame NOT in AspectFrame
        directly

        For AspectFrame to work as intended, pack with fill='both' and
        expand=True, and grid with sticky='nsew'

        DO NOT use AspectFrame.place() because then AspectFrame will not work

        AspectFrame should ALWAYS be packed with expand=True and grid with
        sticky='nsew' and parent widget gridrowconfigure and gridcolumnconfigure
    '''
    def __init__(self, master, aspect_ratio:float, pad_bg:str='#ffffff',
                 frame_bg:str='#ffffff', config_callback:callable=None,
                 size_callback:callable=None):
        '''
        Parameters
        ----------
            :param master: tk.Frame - parent widget
            :param pad_bg: str (hex code) - color of padding
            :param frame_bg: str (hex code) - frame background color
            :param config_callback: function (event) - called whenever frame is resized
            :param size_callback: function (width, height) - called whenever frame is resized
        '''
        Frame.__init__(self, master, bg=pad_bg)
        self.__aspect_ratio = aspect_ratio
        self.__config_callback = config_callback
        self.__size_callback = size_callback

        self.__max_width, self.__max_height = 1, 1 # based on entire frame size
        self.__width, self.__height = 1, 1 # based on aspect ratio within frame

        self.frame = Frame(self, bg=frame_bg)

        self.bind('<Configure>', self.__configure)
        #self.event_generate('<Configure>', when='tail')

    def __configure(self, event):
        '''places self.frame within self with the appropiate aspect ratio'''
        self.__max_width, self.__max_height = event.width, event.height
        self.__width, self.__height = self.__compute_dims()
        
        self.frame.place(in_=self, relx=0.5, rely=0.5, anchor='center',
                         width=self.__width, height=self.__height)

        if self.__config_callback:
            self.__config_callback(event)
        if self.__size_callback:
            self.__size_callback(*self.get_dims())

    def __compute_dims(self):
        '''returns width, height based on max_width, max_height, and aspect_ratio'''
        # start by using the width as the controlling dimension
        width = self.__max_width
        height = int(self.__max_width / self.__aspect_ratio)

        # if window is too tall to fit, use the height as controlling dimension
        if height > self.__max_height:
            height = self.__max_height
            width = int(self.__max_height * self.__aspect_ratio)

        return width, height

    def set_aspect_ratio(self, aspect_ratio:float):
        ''':param aspect_ratio: float - ratio of width / height'''
        self.__aspect_ratio = aspect_ratio
        self.__width, self.__height = self.__compute_dims()
        self.frame.place(in_=self, relx=0.5, rely=0.5, anchor='center',
                         width=self.__width, height=self.__height)

    def get_aspect_ratio(self) -> float:
        '''returns aspect ratio: ratio of width / height'''
        return self.__aspect_ratio
    
    def get_dims(self):
        '''returns inner frame width, height as tuple'''
        return self.__width, self.__height



from tkinter import Frame, Label

from .progress_bar import ProgressBar
from .buttons import IconButton
from .icons import icons


class FunctionProgress(Frame):
    ''' Widget to call a specific function and display the progress of that
        function's execution
        
        Uses IconButton to allow user to start the function
        
        Indictes if function has already been executed

        When function is being executed, it should call self.Progress.increment()
        with each iteration to increment progress bar
    '''
    def __init__(self, master, command, label, bg, loading_text='Loading...',
                 icon_path=None, height=25, indicator_pady=0, font_name='Segoe UI',
                 font_size=10, text_font_size=12, inactive_bg:str='#ffffff',
                 loaded_bg:str='#00ff00', text_side='bottom', active_color='#00ff00',
                 text_color='#cccccc', inactive_color=None):
        '''
        
        Parameters
        ----------
            :param master: tk.Frame - parent widget
            :param command: function () - called when button is clicked
                                        - function should increment progress bar
            :param label: str - label for IconButton - to call command
            :param bg: str (hex code) - widget background color
            :param loading_text: str - text displayed while command is running
            :param icon_path: str - optioal path to png file to replace default 'edit' icon
            :param height: int - height of progress bar
            :param indicator_pady: int - pady for indicator square
            :param inactive_bg: str (hex code) - color of indicator when incomplete
            :param loaded_bg: str (hex code) - color of indicator when complete
            :param text_side: Literal['bottom', 'top', 'right', 'left'] - for progress bar
            :param active_color: str (hex code) - color of active part of progress bar
            :param text_color: str (hex code) - color of text in progress bar
            :param inactive_color: str (hex code) or None - inactive part of progress bar
        '''
        Frame.__init__(self, master, bg=bg)
        self.inactive_bg, self.loaded_bg = inactive_bg, loaded_bg
        inactive_color = inactive_color if inactive_color else bg
        self.complete = False

        self.main_frame = Frame(self, bg=bg)
        self.main_frame.pack(side='top', fill='both', expand=True)
        icon = icon_path if icon_path is not None else icons['edit']
        button = IconButton(self.main_frame, icon, command, label=label,
                            selectable=False, inactive_bg=bg, bar_height=2)
        button.pack(side='left', fill='both', expand=True)
        self.indicator = Label(self.main_frame, text=' ' * 3, bg=self.inactive_bg,
                               font=(font_name, font_size))
        self.indicator.pack(side='right', fill='y', pady=indicator_pady)

        self.Progress = ProgressBar(self, 100, bg, active_color=active_color,
                                    text_color=text_color, inactive_color=inactive_color,
                                    text=loading_text, width=200, height=height,
                                    center=False, text_font_size=text_font_size,
                                    val_font_size=font_size, text_side=text_side)

    def to_progress_bar(self, iterations:int):
        '''displayes progress bar with the specified number of iterations'''
        self.main_frame.pack_forget()
        self.Progress.pack(side='top', fill='both', expand=True)
        self.Progress.reset(iterations=iterations)

    def to_complete(self):
        '''removes progress bar and indicates that function is complete'''
        self.complete = True
        self.Progress.pack_forget()
        self.main_frame.pack(side='top', fill='both', expand=True)
        self.indicator.config(bg=self.loaded_bg)

    def to_incomplete(self):
        '''removes progress bar and indicates that function is incomplete'''
        self.complete = False
        self.Progress.pack_forget()
        self.main_frame.pack(side='top', fill='both', expand=True)
        self.indicator.config(bg=self.inactive_bg)

    def is_complete(self) -> bool:
        '''returns True if function is complete, otherwise False'''
        return self.complete



import cv2, os
import numpy as np
from PIL import Image, ImageTk
from tkinter import Frame, Button, Label

from .tool_tip import ToolTip
from .icons import icons


def hex_to_rgb(hex_code:str):
    '''converts hex code to uint8 rgb'''
    return tuple([int(h, 16) for h in (hex_code[1:3], hex_code[3:5], hex_code[5:])])

def image_replace_colors(img:np.array, colors_list:list):
    '''replaces all white pixels (255, 255, 255) in img with color
        :param colors_list: list[tuple(str, str)] - hex codes'''
    for value, replace in colors_list:
        img[np.all(img == hex_to_rgb(value), axis=-1)] = hex_to_rgb(replace)
    return img


class BaseButton(Frame):
    ''' Parent class for all icon buttons and label buttons
    
        Callback command is called whenever button is clicked. If button is
        selectable, the color of the button will change depending on select
        status.

        BaseButton includes a bar at the bottom of the button which can change
        color based on hover/select status.

        Optionally, a popup will appear when the user hovers the mouse over
        button. The popup is directly above the button by default, but its
        position is adjusted if necessary to keep the popup in view. For
        example, if the button is at the very top of the window, the popup will
        appear beneath the button instead.
    '''
    def __init__(self, master, command, popup_label:str=None, click_popup=None,
                 font_name:str='Segoe UI', font_size:int=10, padx:int=0, pady:int=0,
                 selectable:bool=True, select_on_click:bool=True, selected=False,
                 tool_tip_font_name:str='Segoe UI', tool_tip_font_size:int=10,
                 bar_height:int=0, bar_side='bottom', active_bg:str='#070708',
                 inactive_bg:str='#000000', active_hover_bg=None,
                 inactive_hover_bg=None, active_fg:str='#13ce12',
                 inactive_fg:str='#888888', active_hover_fg:str='#74d573',
                 inactive_hover_fg:str='#74d573', popup_bg=None,
                 active_bar_color=None, inactive_bar_color=None,
                 active_hover_bar_color:str=None,
                 inactive_hover_bar_color:str='#dcdcdc', off_fg:str='#555555'):
        '''inherits from tk.Frame - base button for inheritance from IconButton, LabelButton, ToggleIconButton, etc
        BaseButton only contains bottom bar - everything else must be added by child class
        child class must contain function config_colors

        Parameters
        ----------
            :param master: frame in which to put button
            :param command: function - function to be executed when button is clicked
            :param popup_label: str - text to appear as tool tip when cursor hovers on button
            :param click_popup: str - popup label changes to click_popup temporarily when button is clicked
            :param padx: Int - internal x pad for button
            :param pady: Int - internal y pad for button
            :param selectable: Boolean - if True, button can be selected
            :param bar_height: Int - height of bar at the bottom of button
            :param bar_side: Literal['top', 'bottom', 'left', 'right'] - pack side for bar if bar_height > 0
            
            :param active_bg: str (hex code) - background when button is selected
            :param inactive_bg: str (hex code) - background when button is not selected
            :param active_hover_bg: str (hex code) - background when selected and cursor is hovering - if None: same as active_bg
            :param inactive_hover_bg: str (hex code) - background when not selected and cursor if hovering - if None: same as inactive_bg

            :param active_fg: str (hex code) - icon and label fg when button is selected
            :param inactive_fg: str (hex code) - icon and label fg when button is not selected
            :param active_hover_fg: str (hex code) - icon and label fg when selected and cursor is hovering - if None: same as active_fg
            :param inactive_hover_fg: str (hex code) - icon and label fg when not selected and cursor is hovering - if None: same as inactive_fg

            :param active_bar_color: str (hex code) - bar color when selected - if None: same as active_fg
            :param inactive_bar_color: str (hex code) - bar color when not selected - if None: same as inactive_bg
            :param active_hover_bar_color: str (hex code) - bar color when selected and cursor is hovering - if None: same as active_bar_color
            :param inactive_hover_bar_color: str (hex code) - bar color when not selected and cursor is hovering - if None: same as inactive_bar_color
        '''
        Frame.__init__(self, master)
        self.padx, self.pady = padx, pady
        self.selected, self.hovering = selected, False
        self.active = True # if not active, button will not be responsive
        self.click_command = command
        self.font_name, self.font_size = font_name, font_size
        self.selectable, self.select_on_click = selectable, select_on_click
        self.popup_labels = [popup_label + ' : Inactive' if popup_label else None, popup_label] # either both text or both None
        self.click_popup = click_popup
        self.off_fg = off_fg
        self.bg_colors = [[inactive_bg, inactive_hover_bg if inactive_hover_bg else inactive_bg],
                            [active_bg, active_hover_bg if active_hover_bg else active_bg]]
        self.fg_colors = [[inactive_fg, inactive_hover_fg if inactive_hover_fg else inactive_fg],
                            [active_fg, active_hover_fg if active_hover_fg else active_fg]]
        active_bar_color = active_bar_color if active_bar_color else active_fg
        inactive_bar_color = inactive_bar_color if inactive_bar_color else inactive_bg
        self.bar_colors = [[inactive_bar_color, inactive_hover_bar_color if inactive_hover_bar_color else inactive_bar_color],
                            [active_bar_color, active_hover_bar_color if active_hover_bar_color else active_bar_color]]

        self.tool_tip = ToolTip(self, bg=popup_bg if popup_bg else self.bg_colors[0][0], fg='#ffffff',
                                font=(tool_tip_font_name, tool_tip_font_size))
        self.bar = Frame(self, height=bar_height, width=bar_height)
        if bar_height > 0:
            bar_fills = {'top':'x', 'bottom':'x', 'left':'y', 'right':'y'} # fill depends on bar pack side
            self.bar.pack(side=bar_side, fill=bar_fills[bar_side], expand=True)

        self.bind("<Enter>", self.hover_enter)
        self.bind("<Leave>", self.hover_leave)
        self.bind("<Button-1>", self.click_button)
        self.bar.bind("<Button-1>", self.click_button)

    def config_colors(self):
        '''placeholder - this method must be replaced in inheritance classes'''

    def set_color(self, color:str, which:str='bg', selected:bool=False, hover:bool=False, all:bool=False):
        '''sets a single color
        
        Parameters
        ----------
            :param color: str (hex code)
            :param which: str - options: ['bg', 'fg', 'bar']
            :param selected: bool - selected color
            :param hover: bool - hover color
            :param all bool - if True, changes color for all selected and hover statuses
        '''
        if all:
            for s in [True, False]:
                for h in [True, False]:
                    [self.bg_colors, self.fg_colors, self.bar_colors][['bg', 'fg', 'bar'].index(which)][s][h] = color
        else:
            [self.bg_colors, self.fg_colors, self.bar_colors][['bg', 'fg', 'bar'].index(which)][selected][hover] = color
        self.config_colors()

    def turn_on(self):
        self.active = True
        self.config_colors()

    def turn_off(self):
        self.active = False
        self.config_colors()

    def hover_enter(self, event):
        if not self.hovering:
            self.hovering = True
            if self.popup_labels[0]: # tool tip
                self.tool_tip.fadein(0, self.popup_labels[self.active], event) # 0 is initial alpha
            if self.active:
                self.config_colors()

    def hover_leave(self, event):
        if self.hovering:
            self.hovering = False
            if self.popup_labels[0]: # remove tool tip
                self.tool_tip.fadeout(1, event) # first argument is initial alpha
            if self.active:
                self.config_colors()
            
    def click_button(self, event=None, callback=True):
        if self.active:
            if self.popup_labels[0] and self.click_popup:
                self.tool_tip.set_text(self.click_popup)
            if callback:
                self.click_command()
            if self.select_on_click:
                self.select()

    def set(self, status:bool):
        '''sets selection status, so long as button is selectable'''
        if status:
            self.select()
        else:
            self.deselect()

    def select(self):
        if self.selectable and not self.selected:
            self.selected = True
            self.config_colors()

    def deselect(self):
        if self.selected:
            self.selected = False
            self.config_colors()

    def get(self) -> bool:
        '''returns True if button is selected, otherwise False'''
        return self.selected

    def is_hovering(self) -> bool:
        '''returns True if cursor is hovering on button, otherwise False'''
        return self.hovering

class IconButton(BaseButton):
    ''' Extension of BaseButton with an icon besode the label. The label can
        also be an empty string which results in only an icon.
        
        Given the path to a black and white .png file, the foreground and
        background color (set specifically for each hover/select status) can
        be changed with **kwargs passed to BaseButton
    '''
    def __init__(self, master, icon_path:str, command, label:str='', bar_height:int=3, **kwargs):
        '''
        Parameters
        ----------
            :param master: frame in which to put button
            :param icon_path: str or np.array - path to .png file or image array
            :param click_command: 0 argument function - function to be executed when button is clicked
            :param bar_height: Int - height of bar at the bottom of button
        '''
        BaseButton.__init__(self, master, command, bar_height=bar_height, **kwargs) # bar already packed buttom
        self.icon_frame = Frame(self)
        self.icon_frame.pack(side='top', padx=self.padx, pady=self.pady)
        self.icon = Button(self.icon_frame, borderwidth=0, command=self.click_button)
        self.icon.pack(side='left')
        self.label_text = label
        self.label = Label(self.icon_frame, text=label, font=(self.font_name, self.font_size), bd=0)
        self.label.pack(side='right', fill='y')

        self.icon_frame.bind("<Button-1>", self.click_button)
        self.label.bind("<Button-1>", self.click_button)
        
        # Load icon
        if isinstance(icon_path, str): # path to image
            assert len(icon_path) > 4, f'IconButton Error: Invalid path: {icon_path}'
            assert icon_path[-4:] == '.png', f'IconButton Error: icon_path is not a .png file: {icon_path}'
            assert os.path.exists(icon_path), f'IconButton Error: Path to .png file does not exist: {icon_path}'
            self.base_img = cv2.imread(icon_path)
        elif isinstance(icon_path, np.ndarray): # 3d numpy array
            self.base_img = icon_path
        else:
            raise TypeError(f'Invalid icon input to IconButton: {icon_path}')

        # Create Icons
        self.images = [[None, None], [None, None]]
        for x, y in [[0, 0], [0, 1], [1, 0], [1, 1]]:
            temp_img = image_replace_colors(self.base_img.copy(), [('#ffffff', self.fg_colors[x][y]), ('#000000', self.bg_colors[x][y])])
            self.images[x][y] = ImageTk.PhotoImage(image=Image.fromarray(temp_img), master=self.icon_frame)
        off_img = image_replace_colors(self.base_img.copy(), [('#ffffff', self.off_fg), ('#000000', self.bg_colors[0][0])])
        self.off_icon = ImageTk.PhotoImage(image=Image.fromarray(off_img), master=self.icon_frame)

        self.config_colors()

    def config_colors(self):
        '''sets colors based on self.selected, self.hovering, and self.active'''
        if self.active:
            bg, fg = self.bg_colors[self.selected][self.hovering], self.fg_colors[self.selected][self.hovering]
            bar_color, image = self.bar_colors[self.selected][self.hovering], self.images[self.selected][self.hovering]
        else:
            bg, fg, bar_color, image = self.bg_colors[0][0], self.off_fg, self.bg_colors[0][0], self.off_icon
        self.config(bg=bg)
        self.icon_frame.config(bg=bg)
        self.bar.config(bg=bar_color)
        self.label.config(bg=bg, fg=fg)
        self.icon.config(image=image, bg=bg, activebackground=bg)

    def set_color(self, color:str, which:str='bg', selected:bool=False, hover:bool=False):
        '''sets a single color
        
        color : str (hex code)
        which : str - options: ['bg', 'fg', 'bar']
        selected : bool - selected color
        hover : bool - hover color
        '''
        if which == 'bg':
            self.bg_colors[selected][hover] = color
            temp_img = image_replace_colors(self.base_img.copy(), [('#ffffff', self.fg_colors[selected][hover]), ('#000000', color)])
            self.images[selected][hover] = ImageTk.PhotoImage(image=Image.fromarray(temp_img), master=self.icon_frame)
        elif which == 'fg':
            self.fg_colors[selected][hover] = color
            temp_img = image_replace_colors(self.base_img.copy(), [('#ffffff', color), ('#000000', self.bg_colors[selected][hover])])
            self.images[selected][hover] = ImageTk.PhotoImage(image=Image.fromarray(temp_img), master=self.icon_frame)
        elif which == 'bar':
            self.bar_colors[selected][hover] = color
        self.config_colors()

    def get_label(self) -> str:
        '''returns button label'''
        return self.label_text

class ToggleIconButton(IconButton):
    ''' Toggle version of IconButton. The callback command is given a boolean
        parameter which indicates whether the button is being turned on or off.
    '''
    def __init__(self, master, icon_path:str, command=None, label:str='',
                 bar_height:int=3, **kwargs):
        '''Toggle version of IconButton
        
            :param command: 1 argument function (bool) - True for turn on, False for turn off
        '''
        IconButton.__init__(self, master, icon_path, self.click, label=label,
                            select_on_click=False, bar_height=bar_height,
                            **kwargs)
        self.toggle_command = command

    def click(self):
        '''called when button is clicked'''
        self.selected = not self.selected
        self.config_colors()
        if self.toggle_command is not None:
            self.toggle_command(self.selected)

class DoubleIconButton(Frame):
    ''' DoubleIconButton contains two IconButtons that swap when clicked.
        There are separate commands for the two IconButtons
    '''
    def __init__(self, master, icon1:str, icon2:str, command1, command2,
                 label1:str='', label2:str='', popup_label1=None, popup_label2=None,
                 inactive_fg:str='#888888', active_fg:str='#ffffff', **kwargs):
        '''IconButton that changes when clicked
        
        Parameters
        ----------
            :param master: tk.Frame - parent widget
            :param icon1: str - icon path for first IconButton
            :param icon2: str - icon path for second button
            :param command1: 0 argument function - called when first button is clicked - changes to second button
            :param command2: 0 argument function - called when second button is clicked - changes to first button
        '''
        Frame.__init__(self, master)
        self.__command1, self.__command2 = command1, command2
        self.__double_status = False # False when button 1 is active and True when button 2 is active

        self.Button1 = IconButton(self, icon1, self.click1, label=label1, selectable=False,
                                    popup_label=popup_label1, inactive_fg=inactive_fg, **kwargs)
        self.Button2 = IconButton(self, icon2, self.click2, label=label2, selectable=False,
                                    popup_label=popup_label2, inactive_fg=active_fg, **kwargs)
        self.Button1.pack(fill='both', expand=True)

    def switch1(self):
        '''
        Purpose:
            switches to button1 without calling any command
        Pre-conditions:
            (none)
        Post-conditions:
            switches to button1
        Returns:
            (none)
        '''
        self.__double_status = False
        self.Button2.pack_forget()
        self.Button1.pack(fill='both', expand=True)

    def switch2(self):
        '''
        Purpose:
            switches to button2 without calling any command
        Pre-conditions:
            (none)
        Post-conditions:
            switches to button2
        Returns:
            (none)
        '''
        self.__double_status = True
        self.Button1.pack_forget()
        self.Button2.pack(fill='both', expand=True)

    def click1(self):
        '''
        Purpose:
            called when first button is clicked
            switches to button2 and calls command1
        Pre-conditions:
            (none)
        Post-conditions:
            changes active (visible) button
        Returns:
            (none)
        '''
        self.switch2()
        self.__command1()

    def click2(self):
        '''
        Purpose:
            called when second button is clicked
            switches to button1 and calls command2
        Pre-conditions:
            (none)
        Post-conditions:
            changes active (visible) button
        Returns:
            (none)
        '''
        self.switch1()
        self.__command2()

    def turn_on(self):
        '''makes button interactable'''
        self.Button1.turn_on()
        self.Button2.turn_on()

    def turn_off(self):
        '''makes button uninteractable'''
        self.Button1.turn_off()
        self.Button2.turn_off()

    def set(self, status:bool):
        '''sets to button1 if status is False and button2 if status is True'''
        if status:
            self.switch2()
        else:
            self.switch1()

    def get(self):
        '''returns False if button1 is active and True if button2 is active'''
        return self.__double_status
    
    def get_label(self) -> str:
        '''returns current button label'''
        if self.__double_status: # button2 is active
            return self.Button2.get_label()
        else:
            return self.Button1.get_label()

class CheckButton(DoubleIconButton):
    ''' Special version of DoubleIconButton that has a checkbox icon - checked and unchecked
        Takes a single callback command - passes boolean to indicate new checkbox status
    '''
    def __init__(self, master, command, label:str='', active_popup_label=None,
                 inactive_popup_label=None, active=False, **kwargs):
        '''Button1 is unchecked and Button2 is checked
        when Button1 is clicked, calls command(True) because checkbox is being selected

        Parameters
        ----------
            :param master: tk.Frame - parent widget
            :param command: 1 argument function (bool) called when button is clicked
            :param label: str - text to the left of button
            :param active_popup_label: str or None - hover popup text when button is selected
            :param inactive_popup_label: str or None - hover popup text when button is not selected
            :param active: bool - initial status of CheckButton
        '''
        DoubleIconButton.__init__(self, master, icons['box'], icons['checkbox'],
                                  lambda: command(True), lambda: command(False),
                                  label1=label, label2=label,
                                  popup_label1=inactive_popup_label,
                                  popup_label2=active_popup_label, **kwargs)
        if active:
            self.switch2()

    def select(self):
        '''select check button without calling callback command'''
        self.switch2()

    def deselect(self):
        '''deselect check button without calling callback command'''
        self.switch1()

    def set(self, status:bool):
        '''sets selection status of CheckButton'''
        if status:
            self.select()
        else:
            self.deselect()

class LabelButton(BaseButton):
    ''' LabelButton includes bar underneath the label. The parameter
        'bar_height' can be set to 0 to remove the bar.
        
        Background and foreground color can change based on select/hover status.
    '''
    def __init__(self, master, command, label:str='', **kwargs):
        '''just like IconButton but with no icon'''
        BaseButton.__init__(self, master, command, **kwargs)
        self.label_text = label
        self.label = Label(self, text=label, font=(self.font_name, self.font_size), bd=0)
        self.label.pack(side='top', fill='both', expand=True, padx=self.padx, pady=self.pady)
        self.label.bind("<Button-1>", self.click_button)

        self.config_colors()

    def config_colors(self):
        '''sets colors based on self.selected, self.hovering, and self.active'''
        if self.active:
            bg, fg = self.bg_colors[self.selected][self.hovering], self.fg_colors[self.selected][self.hovering]
            bar_color = self.bar_colors[self.selected][self.hovering]
        else:
            bg, fg, bar_color = self.bg_colors[0][0], self.off_fg, self.bg_colors[0][0]
        self.config(bg=bg)
        self.bar.config(bg=bar_color)
        self.label.config(bg=bg, fg=fg)

    def get_text(self):
        '''returns label text'''
        return self.label_text

    def get_label(self):
        '''duplicates get_text() for compatibility'''
        return self.get_text()

class ToggleLabelButton(LabelButton):
    ''' Extension of LabelButton that passes a boolean argument to callback
        command to indicate whether buttons is being selected or deselected.
    '''
    def __init__(self, master, command=None, label:str='', **kwargs):
        '''Toggle version of LabelButton
        
            :param command: 1 argument function (bool) - True for turn on, False for turn off
        '''
        LabelButton.__init__(self, master, self.click, label=label, select_on_click=False, **kwargs)
        self.toggle_command = command

    def click(self):
        '''called when button is clicked'''
        self.selected = not self.selected
        self.config_colors()
        if self.toggle_command is not None:
            self.toggle_command(self.selected)

class ToggleButtonGroup(Frame):
    ''' Group of ToggleIconButtons where only one can be active at a time

        Depending on specification, it is possible to have a default button
        that is selected if the currently selected button is deselected so that
        there is always exactly one button selected
    
        Clicking a button will automatically deselect all other buttons in the
        group and call their respective off callback functions
    '''
    def __init__(self, master:Frame, button_info:list, callback=None,
                 deselectable=False, always_selected=False, default_index=0,
                 orientation:str='h', bg='#ffffff', buttons_padx=0, buttons_pady=0,
                 **kwargs):
        '''
        Parameters
        ----------
            :master: Frame - parent frame
            :param button_info: list[dict] - contains keys for each button:
                icon_path: str - path to .png file (optional)
                on_callback: function() - called when button is clicked (optional)
                off_callback: function() - called when button is deselected (optional)
                label: str - button label (optional)
                popup_label: str - button popup label (optional)
            :param callback: function(clicked_index) - called when a button is clicked
            :param deselectable: bool - if True, buttons can be deselected by clicking
            :param always_selected: bool - if True, exactly one button will always be selected
            :param default_index: int - index of default button to be selected
        '''
        assert orientation in ['h', 'v'], f'ToggleButtonGroup Error: Invalid orientation: {orientation}'
        super().__init__(master, bg=bg)
        self.__button_info = button_info
        self.__selection_callback = callback
        self.__deselectable = deselectable
        self.__always_selected = always_selected
        self.__default_index = default_index

        pack_side = 'left' if orientation == 'h' else 'top'

        self.__buttons: list[ToggleIconButton] = []
        for i, info in enumerate(self.__button_info):
            label = info['label'] if 'label' in info.keys() else ''
            popup_label = info['popup_label'] if 'popup_label' in info.keys() else ''
            if 'icon_path' in info.keys(): # IconButton
                button = ToggleIconButton(self, info['icon_path'],
                                        command=lambda b, x=i: self.__callback(b, x),
                                        label=label, popup_label=popup_label, **kwargs)
            else: # LabelButton
                button = ToggleLabelButton(self, command=lambda b, x=i: self.__callback(b, x),
                                           label=label, popup_label=popup_label, **kwargs)
            button.pack(side=pack_side, padx=buttons_padx, pady=buttons_pady, fill='both')
            self.__buttons.append(button)

    def __callback(self, active:bool, button_index:int, callback=True):
        '''called when the ith button is clicked'''
        if not active and not self.__deselectable: # disallow deselection
            self.__buttons[button_index].select() # select button that was just deselected
        elif active: # callback and deactivate all other buttons
            for i, button in enumerate(self.__buttons):
                if i == button_index: # callback for clicked button
                    if callback:
                        self.__on_callback(i)
                else: # deactivate other buttons
                    button.deselect()
                    if callback:
                        self.__off_callback(i)
            if self.__selection_callback is not None and callback:
                self.__selection_callback(button_index)
        else: # off callback for clicked button
            if callback:
                self.__off_callback(button_index)
            if self.__always_selected: # select default button
                self.__buttons[self.__default_index].select()
                if callback:
                    self.__on_callback(self.__default_index)
                if self.__selection_callback is not None and callback:
                    self.__selection_callback(button_index)

    def __on_callback(self, i:int):
        '''calls on_callback for the given button index if it exists'''
        if 'on_callback' in self.__button_info[i].keys():
            self.__button_info[i]['on_callback']()

    def __off_callback(self, i:int):
        '''calls off_callback for the given button index if it exists'''
        if 'off_callback' in self.__button_info[i].keys():
            self.__button_info[i]['off_callback']()

    def set_color(self, button_index:int, color:str, **kwargs):
        '''sets color for the button with the specified index'''
        self.__buttons[button_index].set_color(color, **kwargs)

    def click(self, button_index:int):
        '''clicks the specified button'''
        self.__buttons[button_index].click_button()

    def set(self, set_index:int):
        '''selects button with the given index and deselects all others - no callback'''
        self.__buttons[set_index].select()
        for i, button in enumerate(self.__buttons): # deselect other buttons
            if i != set_index:
                button.deselect()

    def get_label(self) -> str:
        '''returns the label of the button that is currently selected
        If none are selected, returns None - only possible when always_selected is False
        '''
        for button in self.__buttons:
            if button.get():
                return button.get_label()

    def get(self) -> list:
        '''returns list of booleans indicating selection status of each button'''
        return [button.get() for button in self.__buttons]

class CheckButtonGroup(Frame):
    ''' Group of CheckButtons where any, none, or all can be selected

        get() returns list of labels of buttons that are selected
    
        CheckButtons can be arranged horizontally, vertically, or in rows and
        columns

        Columns are considered preferentially if rows and columns are defined
    '''
    def __init__(self, master:Frame, labels:list, callback=None, bg:str='#ffffff',
                 orientation:str='h', rows=None, columns=None, selected_labels=None,
                 buttons_padx=0, buttons_pady=0, bottom_pady=10, **kwargs):
        '''
        Parameters
        ----------
            :param master: tk.Frame - parent widget
            :param labels: list[str] - button labels
            :param callback: function(list[str]) - called when selections are changed
            :param bg: str (hex code) - background color
            :param orientation: str Literal['h', 'v'] - horizontal or vertical pack
            :param rows: int or None - number of rows in grid
            :param columns: int or None - number of columns in grid
            :param selected_labels: list[str] - labels to be intially selected
            **kwargs are passed to all CheckButtons
        '''
        assert orientation in ['h', 'v'], f'CheckButtonGroup Error: Invalid orientation: {orientation}'
        super().__init__(master, bg=bg)
        self.__buttons: list[CheckButton] = []
        self.__callback = callback

        pack_side = 'left' if orientation == 'h' else 'top'

        buttons_frame = Frame(self, bg=bg)
        buttons_frame.pack(side='top', fill='both', expand=True)

        # Grid Configure
        if columns is not None:
            buttons_frame.grid_columnconfigure(tuple(range(columns)), weight=1)
        if rows is not None:
            buttons_frame.grid_columnconfigure(tuple(range(rows)), weight=1)

        # Create Buttons
        for i, label in enumerate(labels):
            button = CheckButton(buttons_frame, self.__click, label=label,
                                 active=False, **kwargs)
            self.__buttons.append(button)
            if columns is not None: # grid a row at a time
                row = i // columns
                column = i % columns
                button.grid(row=row, column=column, padx=buttons_padx,
                            pady=buttons_pady, sticky='nsew')
            elif rows is not None: # grid a column at a time
                row = i % rows
                column = i // rows
                button.grid(row=row, column=column, padx=buttons_padx,
                            pady=buttons_pady, sticky='nsew')
            else: # pack according to orientation
                button.pack(side=pack_side, fill='both', expand=True,
                            padx=buttons_padx, pady=buttons_pady)
                
        # Buttons to Select/Deselect All
        bottom_frame = Frame(self, bg=bg)
        bottom_frame.pack(side='bottom', fill='x', pady=bottom_pady)
        bkwargs = dict(bar_height=0, inactive_bg=bg, selectable=False)
        deselect_button = IconButton(bottom_frame, icons['box'], self.deselect_all,
                                     label='None', **bkwargs)
        deselect_button.pack(side='left', fill='x', expand=True)
        select_button = IconButton(bottom_frame, icons['checkbox'], self.select_all,
                                   label='All', **bkwargs)
        select_button.pack(side='right', fill='x', expand=True)

        # Set Initial Selections
        if selected_labels is not None:
            self.set(selected_labels)

    def __click(self, status:bool):
        '''called when any CheckButton is clicked'''
        if self.__callback is not None:
            self.__callback(self.get())

    def select_all(self):
        '''selects all buttons'''
        for button in self.__buttons:
            button.select()
        if self.__callback is not None:
            self.__callback(self.get())

    def deselect_all(self):
        '''deselects all buttons'''
        for button in self.__buttons:
            button.deselect()
        if self.__callback is not None:
            self.__callback(self.get())

    def set(self, labels:list):
        '''selects the given labels and deselects all others'''
        for button in self.__buttons:
            if button.get_label() in labels:
                button.select()
            else:
                button.deselect()

    def get(self) -> list:
        '''returns list of labels for all selected buttons'''
        return [b.get_label() for b in self.__buttons if b.get()]

class PlayerButtons(Frame):
    ''' Collection of IconButtons for controlling the playback of music, a
        video, or something like that.
        
        Includes 6 buttons in the following order, with corresponding callbacks:
            previous - go to previous song or beginning of current song
            skip_back - go back by a given increment
            play/pause - toggle button to start/stop playback
            skip_forward - go forward by a given increment
            next - go to next song or end of current song
            loop (toggle) - turn looping on or off
    '''
    def __init__(self, master, bg, play_function, stop_function, step_forward_function,
                 step_back_function, next_function, previous_function,
                 active_icon_color='#ffffff', hover_fg='#aaaaaa',
                 button_padx=6, padx_weight=6, active=True):
        '''
        Parameters
        ----------
            :param master: tk.Frame - parent widget
            :param bg: str (hex code) - background color
            :param play_function: function () - called when play button is clicked
            :param stop_function: function () - called when stop button is clicked
            :param step_forward_function: function () - called when step_forward button is clicked
            :param step_back_function: function () - called when step_back button is clicked
            :param next_function: function () - called when next button is clicked
            :param previous_function: function () - called when previous button is clicked
            :param active_icon_color: str (hex code) - color of loop button when selected
            :param hover_fg: str (hex code) - color of buttons when cursor is hovering
            :param button_padx: int (pixels) - minimum distance between buttons
            :param padx_weight: int - weight of x padding for button group
            :param active: bool - if True, buttons are interactable by default
        '''
        Frame.__init__(self, master, bg=bg)
        # for x padding so that buttons dont span full width of frame
        self.grid_columnconfigure(0, weight=padx_weight)
        self.grid_columnconfigure(7, weight=padx_weight)
        for i in range(6):
            self.grid_columnconfigure(i + 1, weight=1)
        
        # key-word arguments common to all buttons
        bkwargs = {'bar_height':0, 'inactive_bg':bg,
                   'inactive_hover_fg':hover_fg , 'padx':button_padx}
        
        # Create buttons
        self.buttons: list[IconButton] = []
        previous_button = IconButton(self, icons['skip_back'], previous_function,
                                     popup_label='Skip to Start',
                                     selectable=False, **bkwargs)
        back_button = IconButton(self, icons['replay5'], step_back_function,
                                 popup_label='Back 5 Seconds',
                                 selectable=False, **bkwargs)
        self.play_button = DoubleIconButton(self, icons['play'], icons['pause'],
                                            play_function, stop_function,
                                            popup_label1='Play', popup_label2='Pause',
                                            **bkwargs)
        forward_button = IconButton(self, icons['forward5'], step_forward_function,
                                    popup_label='Forward 5 Seconds',
                                    selectable=False, **bkwargs)
        next_button = IconButton(self, icons['skip_forward'],
                                 next_function, popup_label='Skip to End',
                                 selectable=False, **bkwargs)
        self.loop_button = ToggleIconButton(self, icons['loop'],
                                            popup_label='Toggle Loop',
                                            active_bg=bg, active_fg=active_icon_color,
                                            active_hover_fg=hover_fg, **bkwargs)

        # Grid buttons
        for i, button in enumerate([previous_button, back_button, self.play_button,
                                    forward_button, next_button, self.loop_button]):
            button.grid(row=0, column=i + 1)
            self.buttons.append(button)

        if not active:
            self.turn_off()

    def is_looped(self):
        '''returns True if looping is on, otherwise False'''
        return self.loop_button.selected
    
    def to_stop(self):
        '''switches play button to stop button without calling play command'''
        self.play_button.switch2()

    def to_play(self):
        '''switches stop button to play button without calling stop command'''
        self.play_button.switch1()

    def turn_on(self):
        '''make buttons interactable by user'''
        for button in self.buttons:
            button.turn_on()

    def turn_off(self):
        '''make buttons uninteractable by user'''
        for button in self.buttons:
            button.turn_off()




from tkinter import Toplevel, Frame, Label, Text, Scrollbar, PhotoImage, filedialog
from threading import Thread
import shutil # for copying files
import fitz # fitz is PyMuPDF

from .buttons import IconButton
from .icons import icons


class PdfDisplay(Frame):
    ''' Widget that displays pdf or 'No File Loaded' label when no pdf is loaded
    
        Contains a button that allows the user to open the pdf in a new window
    '''
    def __init__(self, master, bg, fg, font_name:str='Segoe UI', font_size:int=20,
                 button_pad:int=2, view_width=75, zoom_fact=1, height:int=600,
                 buttons_side='left', buttons_bg='#ffffff',
                 new_window_option=True, download_option=True, zoom_options=True):
        '''
        
        Parameters
        ----------
            :param master: tk.Frame - parent widget
            :param bg: str (hex code) - background color
            :param fg: str (hex code) - color of 'No File Loaded' label
            :param font_name: str - font name such as 'Segoe UI'
            :param font_size: int - font size
            :param button_pad: int - padding between button and top left corner
            :param view_width: int - width of view in pixels
            :param zoom_fact: float - factor by which to zoom pdf
            :param buttons_side: str - Literal['left', 'right']
            :param buttons_bg: str (hex code) - background color of buttons
            :param new_window_option: bool - if True, include button to open in new window
            :param download_option: bool - if True, include button to download pdf
            :param zoom_options: bool - if True, include buttons to zoom pdf
        '''
        assert buttons_side in ['left', 'right'], f"Invalid buttons side: '{buttons_side}', must be 'left' or 'right'"
        Frame.__init__(self, master, bg=bg)
        self.button_pad = button_pad
        self.bg = bg # store for new window
        self.width, self.height = view_width, height
        self.zoom_fact = zoom_fact # never changes
        self.scale_fact = 1 # changes when zooming in and out
        self.img_object_list = []
        self.buttons_side = buttons_side
        self.pdf_frame = None # tk.Frame once a pdf is loaded
        self.active = False # True when a pdf is loaded

        # Buttons Frame
        self.buttons_frame = Frame(self, bg='#ffffff')

        bkwargs = {'bar_height':0, 'selectable':False, 'inactive_bg':buttons_bg,
                   'inactive_hover_fg':None, 'popup_bg':self.bg}

        if new_window_option:
            new_button = IconButton(self.buttons_frame, icons['open_in_new'],
                                    command=self.open_in_window,
                                    popup_label='Open In New Window', **bkwargs)
            new_button.pack(side=buttons_side)

        if download_option:
            down_button = IconButton(self.buttons_frame, icons['file_download'],
                                     command=self.download_pdf,
                                     popup_label='Download PDF', **bkwargs)
            down_button.pack(side=buttons_side)

        if zoom_options:
            out_button = IconButton(self.buttons_frame, icons['minus'],
                                    command=self.zoom_out,
                                    popup_label='Zoom Out', **bkwargs)
            in_button = IconButton(self.buttons_frame, icons['plus'],
                                   command=self.zoom_in,
                                   popup_label='Zoom In', **bkwargs)
            out_button.pack(side='left')
            in_button.pack(side='left')
        
        # Inactive Label
        self.label = Label(self, text='No File Loaded', bg=bg, fg=fg,
                           font=(font_name, font_size))
        self.label.pack(fill='both', expand=True)

    def position_buttons(self, event=None):
        '''repositions buttons - called when window is resized'''
        if self.active:
            if self.buttons_side == 'left':
                x = self.pdf_frame.winfo_x() + self.button_pad
                anchor = 'nw'
            elif self.buttons_side == 'right':
                right_edge = self.pdf_frame.winfo_x() + self.pdf_frame.winfo_width()
                x = right_edge - self.scroll_y.winfo_width() - self.button_pad
                anchor = 'ne'
            else: # this should never happen
                raise ValueError(f'PdfDisplay button side is invalid: {self.buttons_side}')
            self.buttons_frame.place(x=x, y=self.pdf_frame.winfo_y() + self.button_pad,
                                     anchor=anchor)
            self.buttons_frame.lift() # raise above pdf

    def remove_all(self):
        '''removes pdf, buttons, and label'''
        self.active = False
        self.label.pack_forget()
        self.buttons_frame.place_forget()
        if self.pdf_frame != None:
            self.pdf_frame.destroy()

    def show_pdf(self, filename:str):
        '''loads pdf from the given filepath'''
        self.scale_fact = 1 # reset zoom
        self.filename = filename
        self.remove_all()
        self.active = True
        Thread(target=self.render_pdf()).start()

    def render_pdf(self):
        '''loads pdf from self.filename - MUST be called in Thread'''
        # Create image objects from pdf pages
        self.img_object_list = []
        for page in fitz.open(self.filename):
            pix = page.get_pixmap(dpi=int(72 * self.zoom_fact * self.scale_fact))
            pix1 = fitz.Pixmap(pix, 0) if pix.alpha else pix
            image = pix1.tobytes("ppm")
            self.img_object_list.append(PhotoImage(data=image))

        # Create new pdf frame
        width = int(self.width * self.scale_fact)
        height = int(self.height * self.scale_fact)
        self.pdf_frame = Frame(self, height=height, width=width, bg=self.bg)
        self.scroll_y = Scrollbar(self.pdf_frame, orient="vertical")
        self.scroll_x = Scrollbar(self.pdf_frame, orient="horizontal")
        self.scroll_x.pack(side='bottom', fill='x')
        self.scroll_y.pack(side='right', fill='y')
        # Font size is used to control the space between pages
        self.pdf_text = Text(self.pdf_frame, bg=self.bg, yscrollcommand=self.scroll_y.set,
                             xscrollcommand=self.scroll_x.set, #font=('Segoe UI', 2),
                             height=height, width=width)
        self.pdf_text.pack(side="left")
        self.scroll_x.config(command=self.pdf_text.xview)
        self.scroll_y.config(command=self.pdf_text.yview)
        self.pdf_frame.pack()
        self.pdf_frame.bind('<Configure>', self.position_buttons)
        self.update() # render so that buttons are positioned properly
        self.position_buttons() # place buttons and raise above pdf

        # Add pages to text box
        for i, img in enumerate(self.img_object_list):
            if i > 0: # add space before next page
                self.pdf_text.insert('end', '\n\n')
            self.pdf_text.image_create('end', image=img)
        self.pdf_text.configure(state='disabled')

    def remove_pdf(self):
        '''removes pdf and goes back to 'No File Loaded' label'''
        self.remove_all()
        self.label.config(text='No File Loaded')
        self.label.pack(fill='both', expand=True)

    def zoom_in(self):
        '''update zoom_fact and re-render pdf'''
        self.scale_fact *= 1.1
        self.remove_all()
        self.active = True
        Thread(target=self.render_pdf()).start()

    def zoom_out(self):
        '''update zoom_fact and re-render pdf'''
        self.scale_fact /= 1.1
        self.remove_all()
        self.active = True
        Thread(target=self.render_pdf()).start()

    def to_loading(self, text='Loading...'):
        '''removes current pdf and displays loading text'''
        self.remove_all()
        self.label.config(text=text)
        self.label.pack(fill='both', expand=True)

    def download_pdf(self):
        '''opens dialog to download the current pdf file
        this can only ever be called when a pdf is being viewed'''
        destination = filedialog.asksaveasfilename(initialdir='/', title='Select destination file',
                                                    filetypes=(('PDF File', '*.pdf'), ('All Files', '*.*')))
        if destination == '': # clicked 'cancel' instead of saving file
            return
        if destination[-4:] != '.pdf':
            destination += '.pdf'
        shutil.copy2(self.filename, destination)

    def open_in_window(self):
        '''opens PDF in new window - can only ever be called when a PDF is being viewed'''
        PdfWindow(self.filename, self.bg)

class PdfWindow(Toplevel):
    ''' Window to view a single PDF File
    
        PDFs being view in PdfWindow cannot be opened in another new window,
        obviously
    '''
    def __init__(self, filepath:str, bg:str, window_title='PDF Viewer',
                 width_fact:float=0.9, height_fact:float=0.99):
        '''creates window to view pdf
        
        Parameters
        ----------
            :param filepath: str - absolute path to a .pdf file
            :param bg: str (hex code) - window background color
            :param window_title: str - window name displayed in top left corner
            :param width_fact: float between 0 and 1 - percentage of screen width covered by window
            :param height_fact: float between 0 and 1 - percentage of screen height covered by window
        '''
        Toplevel.__init__(self)

        self.config(bg=bg)
        self.title(window_title)
        sw, sh = self.winfo_screenwidth(), self.winfo_screenheight()
        w, h = int(sw * width_fact), int(sh * height_fact)
        self.geometry(f'{w}x{h}+{sw // 2 - w // 2}+{sh // 2 - h // 2}')

        pdf = PdfDisplay(self, bg, '#ffffff', view_width=150, zoom_fact=2,
                         new_window_option=False)
        pdf.pack(fill='both', expand=True)
        pdf.show_pdf(filepath)

